import math
import pickle
from contextlib import nullcontext
from itertools import permutations

import numpy
import pytest
from numpy.testing import assert_almost_equal, assert_array_equal

from pyproj import Geod
from pyproj.geod import GeodIntermediateFlag, reverse_azimuth

try:
    from shapely.geometry import (
        LinearRing,
        LineString,
        MultiLineString,
        MultiPoint,
        MultiPolygon,
        Point,
        Polygon,
    )
    from shapely.geometry.polygon import orient

    SHAPELY_LOADED = True
except (ImportError, OSError):
    SHAPELY_LOADED = False


skip_shapely = pytest.mark.skipif(not SHAPELY_LOADED, reason="Missing shapely")


# BOSTON
_BOSTON_LAT = 42.0 + (15.0 / 60.0)
_BOSTON_LON = -71.0 - (7.0 / 60.0)
# PORTLAND
_PORTLAND_LAT = 45.0 + (31.0 / 60.0)
_PORTLAND_LON = -123.0 - (41.0 / 60.0)


@pytest.mark.parametrize("return_back_azimuth", [True, False])
@pytest.mark.parametrize(
    "ellipsoid,true_az12,true_az21,expected_distance",
    [
        ("clrk66", -66.5305947876623, 75.65363415556968, 4164192.708),
        ("WGS84", -66.5305947876623, 75.65363415556968, 4164074.239),
    ],
)
def test_geodesic_inv(
    ellipsoid,
    true_az12,
    true_az21,
    expected_distance,
    return_back_azimuth,
    scalar_and_array,
):
    geod = Geod(ellps=ellipsoid)
    az12, az21, dist = geod.inv(
        scalar_and_array(_BOSTON_LON),
        scalar_and_array(_BOSTON_LAT),
        scalar_and_array(_PORTLAND_LON),
        scalar_and_array(_PORTLAND_LAT),
        return_back_azimuth=return_back_azimuth,
    )
    if not return_back_azimuth:
        az21 = reverse_azimuth(az21)
    assert_almost_equal(
        (az12, az21, dist),
        (
            scalar_and_array(true_az12),
            scalar_and_array(true_az21),
            scalar_and_array(expected_distance),
        ),
        decimal=3,
    )


@pytest.mark.parametrize(
    "lon_start,lat_start,lon_end,lat_end,res12,res21,resdist",
    [
        (_BOSTON_LON, _BOSTON_LAT, _BOSTON_LON, _BOSTON_LAT, 180.0, 0.0, 0.0),
        (
            _BOSTON_LON,
            _BOSTON_LAT,
            -80.79664651607472,
            44.83744724383204,
            -66.53059478766238,
            106.79071710136431,
            832838.5416198927,
        ),
        (
            -80.79664651607472,
            44.83744724383204,
            -91.21816704002396,
            46.536201500764776,
            -73.20928289863558,
            99.32289055927389,
            832838.5416198935,
        ),
        (
            -91.21816704002396,
            46.536201500764776,
            -102.10621593474447,
            47.236494630072166,
            -80.67710944072617,
            91.36325611787134,
            832838.5416198947,
        ),
        (
            -102.10621593474447,
            47.236494630072166,
            -113.06616309750775,
            46.88821539471925,
            -88.63674388212858,
            83.32809401477382,
            832838.5416198922,
        ),
        (
            -113.06616309750775,
            46.88821539471925,
            _PORTLAND_LON,
            _PORTLAND_LAT,
            -96.67190598522616,
            75.65363415556973,
            832838.5416198926,
        ),
    ],
)
def test_geodesic_inv__multiple_points(
    lon_start, lat_start, lon_end, lat_end, res12, res21, resdist, scalar_and_array
):
    geod = Geod(ellps="clrk66")
    o_az12, o_az21, o_dist = geod.inv(
        scalar_and_array(lon_start),
        scalar_and_array(lat_start),
        scalar_and_array(lon_end),
        scalar_and_array(lat_end),
    )
    assert_almost_equal(
        (o_az12, o_az21, o_dist),
        (scalar_and_array(res12), scalar_and_array(res21), scalar_and_array(resdist)),
    )


@pytest.mark.parametrize(
    "flag", [GeodIntermediateFlag.AZIS_DISCARD, GeodIntermediateFlag.AZIS_KEEP]
)
def test_geodesic_inv_intermediate__azis_flag(flag):
    geod = Geod(ellps="clrk66")
    point_count = 7
    res = geod.inv_intermediate(
        lon1=_BOSTON_LON,
        lat1=_BOSTON_LAT,
        lon2=_PORTLAND_LON,
        lat2=_PORTLAND_LAT,
        npts=point_count,
        initial_idx=0,
        terminus_idx=0,
        flags=flag,
        return_back_azimuth=False,
    )
    assert res.npts == point_count
    assert_almost_equal(res.del_s, 694032.1180165777)
    assert_almost_equal(res.dist, 4164192.7080994663)
    assert_almost_equal(
        res.lons,
        [
            _BOSTON_LON,
            -79.12667425528419,
            -87.67579185665936,
            -96.62666097960022,
            -105.7727853838606,
            -114.86985739838673,
            _PORTLAND_LON,
        ],
    )
    assert_almost_equal(
        res.lats,
        [
            _BOSTON_LAT,
            44.46434542986116,
            46.07643938331146,
            47.0159762562249,
            47.23724050253994,
            46.72890509939583,
            _PORTLAND_LAT,
        ],
    )
    if flag == GeodIntermediateFlag.AZIS_DISCARD:
        assert res.azis is None
    else:
        assert_almost_equal(
            res.azis,
            [
                -66.5305947876623,
                -72.03560255472611,
                -78.11540731053181,
                -84.61959820187617,
                -91.32904268899453,
                -97.98697931255812,
                -104.34636584443031,
            ],
        )


@pytest.mark.parametrize(
    "flag", [GeodIntermediateFlag.AZIS_DISCARD, GeodIntermediateFlag.AZIS_KEEP]
)
def test_geodesic_inv_intermediate__azis_flag__numpy(flag):
    geod = Geod(ellps="clrk66")
    point_count = 3
    lons_b = numpy.empty(point_count)
    lats_b = numpy.empty(point_count)
    res = geod.inv_intermediate(
        out_lons=lons_b,
        out_lats=lats_b,
        lon1=_BOSTON_LON,
        lat1=_BOSTON_LAT,
        lon2=_PORTLAND_LON,
        lat2=_PORTLAND_LAT,
        npts=point_count,
        initial_idx=0,
        terminus_idx=0,
        flags=flag,
        return_back_azimuth=False,
    )
    assert res.npts == point_count
    assert_almost_equal(res.del_s, 2082096.3540497331)
    assert_almost_equal(res.dist, 4164192.7080994663)
    assert_almost_equal(res.lons, [_BOSTON_LON, -96.62666098, _PORTLAND_LON])
    assert_almost_equal(
        res.lats,
        [_BOSTON_LAT, 47.01597626, _PORTLAND_LAT],
    )
    if flag == GeodIntermediateFlag.AZIS_DISCARD:
        assert res.azis is None
    else:
        assert_almost_equal(
            res.azis,
            [-66.5305947876623, -84.61959820187617, -104.34636584443031],
        )
    assert res.lons is lons_b
    assert res.lats is lats_b


@pytest.mark.parametrize("return_back_azimuth", [True, False])
def test_geodesic_inv_intermediate__numpy(return_back_azimuth):
    geod = Geod(ellps="clrk66")
    point_count = 5
    lons = numpy.empty(point_count)
    lats = numpy.empty(point_count)
    azis = numpy.empty(point_count)

    with pytest.warns(UserWarning) if return_back_azimuth is None else nullcontext():
        res = geod.inv_intermediate(
            out_lons=lons,
            out_lats=lats,
            out_azis=azis,
            lon1=_BOSTON_LON,
            lat1=_BOSTON_LAT,
            lon2=_PORTLAND_LON,
            lat2=_PORTLAND_LAT,
            npts=point_count,
            initial_idx=0,
            terminus_idx=0,
            return_back_azimuth=return_back_azimuth,
        )
    assert res.npts == point_count
    assert_almost_equal(res.del_s, 1041048.1770248666)
    assert_almost_equal(res.dist, 4164192.7080994663)
    assert_almost_equal(
        res.lons,
        [_BOSTON_LON, -83.34061499, -96.62666098, -110.34292364, _PORTLAND_LON],
    )
    assert_almost_equal(
        res.lats, [_BOSTON_LAT, 45.35049848, 47.01597626, 47.07350417, _PORTLAND_LAT]
    )
    out_azis = res.azis[:-1]
    if return_back_azimuth in [True, None]:
        out_azis = reverse_azimuth(out_azis)
    assert_almost_equal(
        out_azis, [-66.53059479, -75.01125433, -84.6195982, -94.68069764]
    )
    assert res.lons is lons
    assert res.lats is lats
    assert res.azis is azis


@pytest.mark.parametrize(
    "del_s_fact,flag",
    [
        (1, GeodIntermediateFlag.NPTS_ROUND),
        (4.5 / 5, GeodIntermediateFlag.NPTS_TRUNC),
        (5.5 / 5, GeodIntermediateFlag.NPTS_CEIL),
    ],
)
def test_geodesic_inv_intermediate__del_s__numpy(del_s_fact, flag):
    geod = Geod(ellps="clrk66")
    point_count = 5
    lons = numpy.empty(point_count)
    lats = numpy.empty(point_count)
    azis = numpy.empty(point_count)
    dist = 4164192.7080994663
    del_s = dist / (point_count - 1)
    res = geod.inv_intermediate(
        out_lons=lons,
        out_lats=lats,
        out_azis=azis,
        lon1=_BOSTON_LON,
        lat1=_BOSTON_LAT,
        lon2=_PORTLAND_LON,
        lat2=_PORTLAND_LAT,
        del_s=del_s * del_s_fact,
        initial_idx=0,
        terminus_idx=0,
        flags=flag,
        return_back_azimuth=False,
    )
    assert res.npts == point_count
    assert_almost_equal(res.del_s, del_s)
    assert_almost_equal(res.dist, dist)
    assert_almost_equal(
        res.lons,
        [_BOSTON_LON, -83.34061499, -96.62666098, -110.34292364, _PORTLAND_LON],
    )
    assert_almost_equal(
        res.lats, [_BOSTON_LAT, 45.35049848, 47.01597626, 47.07350417, _PORTLAND_LAT]
    )
    assert_almost_equal(
        res.azis[:-1], [-66.53059479, -75.01125433, -84.6195982, -94.68069764]
    )
    assert res.lons is lons
    assert res.lats is lats
    assert res.azis is azis


@pytest.mark.parametrize("return_back_azimuth", [True, False])
def test_geodesic_fwd_intermediate__numpy(return_back_azimuth):
    geod = Geod(ellps="clrk66")
    point_count = 5
    lons = numpy.empty(point_count)
    lats = numpy.empty(point_count)
    azis = numpy.empty(point_count)
    true_az12 = -66.5305947876623
    dist = 4164192.7080994663
    del_s = dist / (point_count - 1)
    with pytest.warns(UserWarning) if return_back_azimuth is None else nullcontext():
        res = geod.fwd_intermediate(
            out_lons=lons,
            out_lats=lats,
            out_azis=azis,
            lon1=_BOSTON_LON,
            lat1=_BOSTON_LAT,
            azi1=true_az12,
            npts=point_count,
            del_s=del_s,
            initial_idx=0,
            terminus_idx=0,
            return_back_azimuth=return_back_azimuth,
        )
    assert res.npts == point_count
    assert res.lons is lons
    assert res.lats is lats
    assert res.azis is azis
    assert_almost_equal(res.del_s, del_s)
    assert_almost_equal(res.dist, dist)
    assert_almost_equal(
        res.lons,
        [-71.11666667, -83.34061499, -96.62666098, -110.34292364, -123.68333333],
    )
    assert_almost_equal(
        res.lats, [42.25, 45.35049848, 47.01597626, 47.07350417, 45.51666667]
    )
    if return_back_azimuth in [True, None]:
        azis = reverse_azimuth(azis)
    assert_almost_equal(
        azis, [-66.53059479, -75.01125433, -84.6195982, -94.68069764, -104.34636584]
    )


@pytest.mark.parametrize("return_back_azimuth", [True, False])
@pytest.mark.parametrize(
    "ellipsoid,true_az12,true_az21,expected_distance",
    [
        ("clrk66", -66.5305947876623, 75.65363415556968, 4164192.708),
        ("WGS84", -66.5305947876623, 75.65363415556968, 4164074.239),
    ],
)
def test_geodesic_fwd(
    ellipsoid,
    true_az12,
    true_az21,
    expected_distance,
    return_back_azimuth,
    scalar_and_array,
):
    geod = Geod(ellps=ellipsoid)
    endlon, endlat, backaz = geod.fwd(
        scalar_and_array(_BOSTON_LON),
        scalar_and_array(_BOSTON_LAT),
        scalar_and_array(true_az12),
        scalar_and_array(expected_distance),
        return_back_azimuth=return_back_azimuth,
    )
    if not return_back_azimuth:
        backaz = reverse_azimuth(backaz)
    assert_almost_equal(
        (endlon, endlat, backaz),
        (
            scalar_and_array(_PORTLAND_LON),
            scalar_and_array(_PORTLAND_LAT),
            scalar_and_array(true_az21),
        ),
        decimal=3,
    )


@pytest.mark.parametrize("include_initial", [True, False])
@pytest.mark.parametrize("include_terminus", [True, False])
def test_geodesic_npts(include_initial, include_terminus):
    geod = Geod(ellps="clrk66")
    initial_idx = int(not include_initial)
    terminus_idx = int(not include_terminus)
    lonlats = geod.npts(
        _BOSTON_LON,
        _BOSTON_LAT,
        _PORTLAND_LON,
        _PORTLAND_LAT,
        npts=6 - initial_idx - terminus_idx,
        initial_idx=initial_idx,
        terminus_idx=terminus_idx,
    )
    expected_lonlats = [
        (-80.797, 44.837),
        (-91.218, 46.536),
        (-102.106, 47.236),
        (-113.066, 46.888),
    ]
    if include_initial:
        expected_lonlats.insert(0, (_BOSTON_LON, _BOSTON_LAT))
    if include_terminus:
        expected_lonlats.append((_PORTLAND_LON, _PORTLAND_LAT))
    assert_almost_equal(lonlats, expected_lonlats, decimal=3)


@pytest.mark.parametrize(
    "input_data",
    [
        [1.0, 2.0, 3.0, math.nan],
        [1.0, 2.0, math.nan, 4.0],
        [1.0, math.nan, 3.0, 4.0],
        [math.nan, 2.0, 3.0, 4.0],
        [math.nan, math.nan, math.nan, math.nan],
    ],
)
def test_geodesic_npts__nan(input_data):
    geod = Geod(ellps="WGS84")
    assert_array_equal(geod.npts(*input_data, npts=1), [(math.nan, math.nan)])


@pytest.mark.parametrize(
    "ellipsoid,expected_azi12,expected_az21,expected_dist",
    [("clrk66", -66.531, 75.654, 4164192.708), ("WGS84", -66.530, 75.654, 4164074.239)],
)
def test_geodesic_inv__pickle(
    ellipsoid, expected_azi12, expected_az21, expected_dist, tmp_path, scalar_and_array
):
    geod = Geod(ellps=ellipsoid)
    az12, az21, dist = geod.inv(
        scalar_and_array(_BOSTON_LON),
        scalar_and_array(_BOSTON_LAT),
        scalar_and_array(_PORTLAND_LON),
        scalar_and_array(_PORTLAND_LAT),
    )
    assert_almost_equal(
        (az12, az21, dist),
        (
            scalar_and_array(expected_azi12),
            scalar_and_array(expected_az21),
            scalar_and_array(expected_dist),
        ),
        decimal=3,
    )
    pickle_file = tmp_path / "geod1.pickle"
    with open(pickle_file, "wb") as gp1w:
        pickle.dump(geod, gp1w, -1)
    with open(pickle_file, "rb") as gp1:
        geod_pickle = pickle.load(gp1)
    pickle_az12, pickle_az21, pickle_dist = geod_pickle.inv(
        scalar_and_array(_BOSTON_LON),
        scalar_and_array(_BOSTON_LAT),
        scalar_and_array(_PORTLAND_LON),
        scalar_and_array(_PORTLAND_LAT),
    )
    assert_almost_equal(
        (pickle_az12, pickle_az21, pickle_dist),
        (
            scalar_and_array(expected_azi12),
            scalar_and_array(expected_az21),
            scalar_and_array(expected_dist),
        ),
        decimal=3,
    )


def test_geodesic_inv__string_init(scalar_and_array):
    geod = Geod("+ellps=clrk66")
    az12, az21, dist = geod.inv(
        scalar_and_array(_BOSTON_LON),
        scalar_and_array(_BOSTON_LAT),
        scalar_and_array(_PORTLAND_LON),
        scalar_and_array(_PORTLAND_LAT),
    )
    assert_almost_equal(
        (az12, az21, dist),
        (
            scalar_and_array(-66.531),
            scalar_and_array(75.654),
            scalar_and_array(4164192.708),
        ),
        decimal=3,
    )


def test_line_length__single_point():
    geod = Geod(ellps="WGS84")
    assert geod.line_length(1, 1) == 0


def test_line_length__radians():
    geod = Geod(ellps="WGS84")
    total_length = geod.line_length([1, 2], [0.5, 1], radians=True)
    assert_almost_equal(total_length, 5426061.32197463, decimal=3)


def test_line_lengths__single_point():
    geod = Geod(ellps="WGS84")
    assert geod.line_lengths(1, 1) == 0


def test_line_lengths__radians():
    geod = Geod(ellps="WGS84")
    line_lengths = geod.line_lengths([1, 2], [0.5, 1], radians=True)
    assert_almost_equal(line_lengths, [5426061.32197463], decimal=3)


def test_polygon_area_perimeter__single_point():
    geod = Geod(ellps="WGS84")
    area, perimeter = geod.polygon_area_perimeter(1, 1)
    assert area == 0
    assert perimeter == 0


@skip_shapely
def test_geometry_length__point():
    geod = Geod(ellps="WGS84")
    assert geod.geometry_length(Point(1, 2)) == 0


@skip_shapely
def test_geometry_length__linestring():
    geod = Geod(ellps="WGS84")
    assert_almost_equal(
        geod.geometry_length(LineString([Point(1, 2), Point(3, 4)])),
        313588.39721259556,
        decimal=2,
    )


@skip_shapely
def test_geometry_length__linestring__radians():
    geod = Geod(ellps="WGS84")
    assert_almost_equal(
        geod.geometry_length(
            LineString(
                [
                    Point(math.radians(1), math.radians(2)),
                    Point(math.radians(3), math.radians(4)),
                ]
            ),
            radians=True,
        ),
        313588.39721259556,
        decimal=2,
    )


@skip_shapely
def test_geometry_length__linearring():
    geod = Geod(ellps="WGS84")
    assert_almost_equal(
        geod.geometry_length(
            LinearRing(LineString([Point(1, 2), Point(3, 4), Point(5, 2)]))
        ),
        1072185.2103813463,
        decimal=2,
    )


@skip_shapely
def test_geometry_length__polygon():
    geod = Geod(ellps="WGS84")
    assert_almost_equal(
        geod.geometry_length(
            Polygon(LineString([Point(1, 2), Point(3, 4), Point(5, 2)]))
        ),
        1072185.2103813463,
        decimal=2,
    )


@skip_shapely
def test_geometry_length__polygon__radians():
    geod = Geod(ellps="WGS84")
    assert_almost_equal(
        geod.geometry_length(
            Polygon(
                LineString(
                    [
                        Point(math.radians(1), math.radians(2)),
                        Point(math.radians(3), math.radians(4)),
                        Point(math.radians(5), math.radians(2)),
                    ]
                )
            ),
            radians=True,
        ),
        1072185.2103813463,
        decimal=2,
    )


@skip_shapely
def test_geometry_length__multipolygon():
    geod = Geod(ellps="WGS84")
    polygon = Polygon(LineString([Point(1, 2), Point(3, 4), Point(5, 2)]))
    assert_almost_equal(
        geod.geometry_length(MultiPolygon([polygon, polygon])),
        2 * 1072185.2103813463,
        decimal=2,
    )


@skip_shapely
def test_geometry_length__multipolygon__radians():
    geod = Geod(ellps="WGS84")
    polygon = Polygon(
        LineString(
            [
                Point(math.radians(1), math.radians(2)),
                Point(math.radians(3), math.radians(4)),
                Point(math.radians(5), math.radians(2)),
            ]
        )
    )
    assert_almost_equal(
        geod.geometry_length(MultiPolygon([polygon, polygon]), radians=True),
        2 * 1072185.2103813463,
        decimal=2,
    )


@skip_shapely
def test_geometry_length__multilinestring():
    geod = Geod(ellps="WGS84")
    line_string = LineString([Point(1, 2), Point(3, 4), Point(5, 2)])
    assert_almost_equal(
        geod.geometry_length(MultiLineString([line_string, line_string])),
        1254353.5888503822,
        decimal=2,
    )


@skip_shapely
def test_geometry_length__multipoint():
    geod = Geod(ellps="WGS84")
    assert (
        geod.geometry_length(MultiPoint([Point(1, 2), Point(3, 4), Point(5, 2)])) == 0
    )


@skip_shapely
def test_geometry_area_perimeter__point():
    geod = Geod(ellps="WGS84")
    assert geod.geometry_area_perimeter(Point(1, 2)) == (0, 0)


@skip_shapely
def test_geometry_area_perimeter__linestring():
    geod = Geod(ellps="WGS84")
    assert_almost_equal(
        geod.geometry_area_perimeter(LineString([Point(1, 2), Point(3, 4)])),
        (0.0, 627176.7944251911),
        decimal=2,
    )


@skip_shapely
def test_geometry_area_perimeter__linestring__radians():
    geod = Geod(ellps="WGS84")
    assert_almost_equal(
        geod.geometry_area_perimeter(
            LineString(
                [
                    Point(math.radians(1), math.radians(2)),
                    Point(math.radians(3), math.radians(4)),
                ]
            ),
            radians=True,
        ),
        (0.0, 627176.7944251911),
        decimal=2,
    )


@skip_shapely
def test_geometry_area_perimeter__linearring():
    geod = Geod(ellps="WGS84")
    assert_almost_equal(
        geod.geometry_area_perimeter(
            LinearRing(LineString([Point(1, 2), Point(3, 4), Point(5, 2)]))
        ),
        (-49187690467.58623, 1072185.2103813463),
        decimal=2,
    )


@skip_shapely
def test_geometry_area_perimeter__polygon():
    geod = Geod(ellps="WGS84")
    assert_almost_equal(
        geod.geometry_area_perimeter(
            Polygon(LineString([Point(1, 2), Point(3, 4), Point(5, 2)]))
        ),
        (-49187690467.58623, 1072185.2103813463),
        decimal=2,
    )


@skip_shapely
def test_geometry_area_perimeter__polygon__radians():
    geod = Geod(ellps="WGS84")
    assert_almost_equal(
        geod.geometry_area_perimeter(
            Polygon(
                LineString(
                    [
                        Point(math.radians(1), math.radians(2)),
                        Point(math.radians(3), math.radians(4)),
                        Point(math.radians(5), math.radians(2)),
                    ]
                )
            ),
            radians=True,
        ),
        (-49187690467.58623, 1072185.2103813463),
        decimal=2,
    )


@skip_shapely
def test_geometry_area_perimeter__polygon__holes():
    geod = Geod(ellps="WGS84")

    polygon = Polygon(
        LineString([Point(1, 1), Point(1, 10), Point(10, 10), Point(10, 1)]),
        holes=[LineString([Point(1, 2), Point(3, 4), Point(5, 2)])],
    )

    assert_almost_equal(
        geod.geometry_area_perimeter(orient(polygon, 1)),
        (944373881400.3394, 3979008.0359657984),
        decimal=2,
    )

    assert_almost_equal(
        geod.geometry_area_perimeter(orient(polygon, -1)),
        (-944373881400.3394, 3979008.0359657984),
        decimal=2,
    )


@skip_shapely
def test_geometry_area_perimeter__multipolygon():
    geod = Geod(ellps="WGS84")
    polygon = Polygon(LineString([Point(1, 2), Point(3, 4), Point(5, 2)]))
    assert_almost_equal(
        geod.geometry_area_perimeter(MultiPolygon([polygon, polygon])),
        (-98375380935.17245, 2144370.4207626926),
        decimal=2,
    )


@skip_shapely
def test_geometry_area_perimeter__multipolygon__radians():
    geod = Geod(ellps="WGS84")
    polygon = Polygon(
        LineString(
            [
                Point(math.radians(1), math.radians(2)),
                Point(math.radians(3), math.radians(4)),
                Point(math.radians(5), math.radians(2)),
            ]
        )
    )
    assert_almost_equal(
        geod.geometry_area_perimeter(MultiPolygon([polygon, polygon]), radians=True),
        (-98375380935.17245, 2144370.4207626926),
        decimal=2,
    )


@skip_shapely
def test_geometry_area_perimeter__multilinestring():
    geod = Geod(ellps="WGS84")
    line_string = LineString([Point(1, 2), Point(3, 4), Point(5, 2)])
    assert_almost_equal(
        geod.geometry_area_perimeter(MultiLineString([line_string, line_string])),
        (-98375380935.17245, 2144370.4207626926),
        decimal=2,
    )


@skip_shapely
def test_geometry_area_perimeter__multipoint():
    geod = Geod(ellps="WGS84")
    assert geod.geometry_area_perimeter(
        MultiPoint([Point(1, 2), Point(3, 4), Point(5, 2)])
    ) == (0, 0)


@pytest.mark.parametrize(
    "lon,lat,az", permutations([10.0, [10.0], (10.0,)])
)  # 6 test cases
def test_geod_fwd_honours_input_types(lon, lat, az):
    # 622
    gg = Geod(ellps="clrk66")
    outx, outy, outz = gg.fwd(lons=lon, lats=lat, az=az, dist=0)
    assert isinstance(outx, type(lon))
    assert isinstance(outy, type(lat))
    assert isinstance(outz, type(az))


def test_geod_fwd_radians():
    g = Geod(ellps="clrk66")
    lon1 = 1
    lat1 = 1
    az1 = 1
    dist = 1
    assert_almost_equal(
        numpy.rad2deg(g.fwd(lon1, lat1, az1, dist, radians=True)),
        g.fwd(lon1 * 180 / numpy.pi, lat1 * 180 / numpy.pi, az1 * 180 / numpy.pi, dist),
    )


def test_geod_inv_radians():
    g = Geod(ellps="clrk66")
    lon1 = 0
    lat1 = 0
    lon2 = 1
    lat2 = 1
    # the third output is in distance, so we don't want to change from deg-rad there
    out_rad = list(g.inv(lon1, lat1, lon2, lat2, radians=True))
    out_rad[0] *= 180 / numpy.pi
    out_rad[1] *= 180 / numpy.pi
    assert_almost_equal(
        out_rad,
        g.inv(
            lon1 * 180 / numpy.pi,
            lat1 * 180 / numpy.pi,
            lon2 * 180 / numpy.pi,
            lat2 * 180 / numpy.pi,
        ),
    )


@pytest.mark.parametrize("func_name", ("fwd", "inv"))
@pytest.mark.parametrize("radians", (True, False))
def test_geod_scalar_array(func_name, radians):
    # verify two singlepoint calculations match an array of length two
    g = Geod(ellps="clrk66")
    func = getattr(g, func_name)
    assert_almost_equal(
        numpy.transpose([func(0, 0, 1, 1, radians=radians) for i in range(2)]),
        func([0, 0], [0, 0], [1, 1], [1, 1], radians=radians),
    )


@pytest.mark.parametrize(
    "lons1,lats1,lons2", permutations([10.0, [10.0], (10.0,), numpy.array([10.0])], 3)
)  # 6 test cases
def test_geod_inv_honours_input_types(lons1, lats1, lons2):
    # 622
    gg = Geod(ellps="clrk66")
    outx, outy, outz = gg.inv(lons1=lons1, lats1=lats1, lons2=lons2, lats2=0)
    assert isinstance(outx, type(lons1))
    assert isinstance(outy, type(lats1))
    assert isinstance(outz, type(lons2))


def test_geodesic_fwd_inv_inplace():
    gg = Geod(ellps="clrk66")
    _BOSTON_LON = numpy.array([0], dtype=numpy.float64)
    _BOSTON_LAT = numpy.array([0], dtype=numpy.float64)
    _PORTLAND_LON = numpy.array([1], dtype=numpy.float64)
    _PORTLAND_LAT = numpy.array([1], dtype=numpy.float64)

    az12, az21, dist = gg.inv(
        _BOSTON_LON, _BOSTON_LAT, _PORTLAND_LON, _PORTLAND_LAT, inplace=True
    )
    assert az12 is _BOSTON_LON
    assert az21 is _BOSTON_LAT
    assert dist is _PORTLAND_LON

    endlon, endlat, backaz = gg.fwd(_BOSTON_LON, _BOSTON_LAT, az12, dist, inplace=True)
    assert endlon is _BOSTON_LON
    assert endlat is _BOSTON_LAT
    assert backaz is az12


@pytest.mark.parametrize("kwarg", ["b", "f", "es", "rf", "e"])
def test_geod__build_kwargs(kwarg):
    gg = Geod(ellps="clrk66")
    if kwarg == "rf":
        value = 1.0 / gg.f
    elif kwarg == "e":
        value = math.sqrt(gg.es)
    else:
        value = getattr(gg, kwarg)
    gg2 = Geod(a=gg.a, **{kwarg: value})
    assert_almost_equal(gg.a, gg2.a)
    assert_almost_equal(gg.b, gg2.b)
    assert_almost_equal(gg.f, gg2.f)
    assert_almost_equal(gg.es, gg2.es)


@pytest.mark.parametrize("radians", [False, True])
def test_geod__reverse_azimuth(radians):
    f = math.pi / 180 if radians else 1
    xy = numpy.array(
        [
            [0, 0 + 180],
            [180, 180 - 180],
            [-180, -180 + 180],
            [10, 10 - 180],
            [20, 20 - 180],
            [-10, -10 + 180],
        ]
    )
    for x, y in xy:
        assert_almost_equal(reverse_azimuth(x * f, radians=radians), y * f)

    xx = xy.T[0]
    yy = xy.T[1]
    assert_almost_equal(reverse_azimuth(xx * f, radians=radians), yy * f)



include "proj.pxi"


from pyproj.enums import WktVersion

from cpython cimport bool


cdef extern from "proj_experimental.h":
    PJ *proj_crs_promote_to_3D(PJ_CONTEXT *ctx,
                               const char* crs_3D_name,
                               const PJ* crs_2D)

    PJ *proj_crs_demote_to_2D(PJ_CONTEXT *ctx,
                              const char *crs_2D_name,
                              const PJ *crs_3D)

cdef tuple _get_concatenated_operations(PJ_CONTEXT*context, PJ*concatenated_operation)
cdef _to_proj4(
    PJ_CONTEXT* context,
    PJ* projobj,
    object version,
    bint pretty,
)
cdef _to_wkt(
    PJ_CONTEXT* context,
    PJ* projobj,
    object version,
    bint pretty,
    bool output_axis_rule=*,
)

cdef class Axis:
    cdef readonly str name
    cdef readonly str abbrev
    cdef readonly str direction
    cdef readonly double unit_conversion_factor
    cdef readonly str unit_name
    cdef readonly str unit_auth_code
    cdef readonly str unit_code

    @staticmethod
    cdef Axis create(PJ_CONTEXT* context, PJ* projobj, int index)

cdef create_area_of_use(PJ_CONTEXT* context, PJ* projobj)

cdef class Base:
    cdef PJ *projobj
    cdef PJ_CONTEXT* context
    cdef readonly str name
    cdef readonly str _remarks
    cdef readonly str _scope
    cdef _set_base_info(self)

cdef class _CRSParts(Base):
    pass


cdef class Ellipsoid(_CRSParts):
    cdef readonly double semi_major_metre
    cdef readonly double semi_minor_metre
    cdef readonly bint is_semi_minor_computed
    cdef readonly double inverse_flattening

    @staticmethod
    cdef Ellipsoid create(PJ_CONTEXT* context, PJ* ellipsoid_pj)


cdef class PrimeMeridian(_CRSParts):
    cdef readonly double longitude
    cdef readonly double unit_conversion_factor
    cdef readonly str unit_name

    @staticmethod
    cdef PrimeMeridian create(PJ_CONTEXT* context, PJ* prime_meridian_pj)


cdef class Datum(_CRSParts):
    cdef readonly str type_name
    cdef readonly object _ellipsoid
    cdef readonly object _prime_meridian

    @staticmethod
    cdef Datum create(PJ_CONTEXT* context, PJ* datum_pj)


cdef class CoordinateSystem(_CRSParts):
    cdef readonly list _axis_list

    @staticmethod
    cdef CoordinateSystem create(PJ_CONTEXT* context, PJ* coordinate_system_pj)


cdef class Param:
    cdef readonly str name
    cdef readonly str auth_name
    cdef readonly str code
    cdef readonly object value
    cdef readonly double unit_conversion_factor
    cdef readonly str unit_name
    cdef readonly str unit_auth_name
    cdef readonly str unit_code
    cdef readonly str unit_category

    @staticmethod
    cdef Param create(PJ_CONTEXT* context, PJ* projobj, int param_idx)


cdef class Grid:
    cdef readonly str short_name
    cdef readonly str full_name
    cdef readonly str package_name
    cdef readonly str url
    cdef readonly bint direct_download
    cdef readonly bint open_license
    cdef readonly bint available

    @staticmethod
    cdef Grid create(PJ_CONTEXT* context, PJ* projobj, int grid_idx)


cdef class CoordinateOperation(_CRSParts):
    cdef readonly list _params
    cdef readonly list _grids
    cdef readonly object _area_of_use
    cdef readonly str method_name
    cdef readonly str method_auth_name
    cdef readonly str method_code
    cdef readonly double accuracy
    cdef readonly bint is_instantiable
    cdef readonly bint has_ballpark_transformation
    cdef readonly list _towgs84
    cdef readonly tuple _operations
    cdef readonly str type_name

    @staticmethod
    cdef CoordinateOperation create(PJ_CONTEXT* context, PJ* coordinate_operation_pj)


cdef class _CRS(Base):
    cdef PJ_TYPE _type
    cdef PJ_PROJ_INFO projpj_info
    cdef readonly str srs
    cdef readonly str _type_name
    cdef readonly Ellipsoid _ellipsoid
    cdef readonly object _area_of_use
    cdef readonly PrimeMeridian _prime_meridian
    cdef readonly Datum _datum
    cdef readonly list _sub_crs_list
    cdef readonly _CRS _source_crs
    cdef readonly _CRS _target_crs
    cdef readonly _CRS _geodetic_crs
    cdef readonly CoordinateSystem _coordinate_system
    cdef readonly CoordinateOperation _coordinate_operation



include "proj.pxi"

cpdef str _get_proj_error() noexcept
cpdef void _clear_proj_error() noexcept
cdef PJ_CONTEXT* PYPROJ_GLOBAL_CONTEXT
cdef PJ_CONTEXT* pyproj_context_create() except *
cdef void pyproj_context_destroy(PJ_CONTEXT* context) except *



"""
This module is for building operations to be used when
building a CRS.

:ref:`operations`
"""
# pylint: disable=too-many-lines
import warnings
from typing import Any

from pyproj._crs import CoordinateOperation
from pyproj.exceptions import CRSError


class AlbersEqualAreaConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Albers Equal Area Conversion.

    :ref:`PROJ docs <aea>`
    """

    def __new__(
        cls,
        latitude_first_parallel: float,
        latitude_second_parallel: float,
        latitude_false_origin: float = 0.0,
        longitude_false_origin: float = 0.0,
        easting_false_origin: float = 0.0,
        northing_false_origin: float = 0.0,
    ):
        """
        Parameters
        ----------
        latitude_first_parallel: float
            First standard parallel (lat_1).
        latitude_second_parallel: float
            Second standard parallel (lat_2).
        latitude_false_origin: float, default=0.0
            Latitude of projection center (lat_0).
        longitude_false_origin: float, default=0.0
            Longitude of projection center (lon_0).
        easting_false_origin: float, default=0.0
            False easting (x_0).
        northing_false_origin: float, default=0.0
            False northing (y_0).
        """
        aea_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {
                "name": "Albers Equal Area",
                "id": {"authority": "EPSG", "code": 9822},
            },
            "parameters": [
                {
                    "name": "Latitude of false origin",
                    "value": latitude_false_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8821},
                },
                {
                    "name": "Longitude of false origin",
                    "value": longitude_false_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8822},
                },
                {
                    "name": "Latitude of 1st standard parallel",
                    "value": latitude_first_parallel,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8823},
                },
                {
                    "name": "Latitude of 2nd standard parallel",
                    "value": latitude_second_parallel,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8824},
                },
                {
                    "name": "Easting at false origin",
                    "value": easting_false_origin,
                    "unit": {
                        "type": "LinearUnit",
                        "name": "Metre",
                        "conversion_factor": 1,
                    },
                    "id": {"authority": "EPSG", "code": 8826},
                },
                {
                    "name": "Northing at false origin",
                    "value": northing_false_origin,
                    "unit": {
                        "type": "LinearUnit",
                        "name": "Metre",
                        "conversion_factor": 1,
                    },
                    "id": {"authority": "EPSG", "code": 8827},
                },
            ],
        }
        return cls.from_json_dict(aea_json)


class AzimuthalEquidistantConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0 AzumuthalEquidistantConversion
    .. versionadded:: 3.2.0 AzimuthalEquidistantConversion

    Class for constructing the Modified Azimuthal Equidistant conversion.

    :ref:`PROJ docs <aeqd>`
    """

    def __new__(
        cls,
        latitude_natural_origin: float = 0.0,
        longitude_natural_origin: float = 0.0,
        false_easting: float = 0.0,
        false_northing: float = 0.0,
    ):
        """
        Parameters
        ----------
        latitude_natural_origin: float, default=0.0
            Latitude of projection center (lat_0).
        longitude_natural_origin: float, default=0.0
            Longitude of projection center (lon_0).
        false_easting: float, default=0.0
            False easting (x_0).
        false_northing: float, default=0.0
            False northing (y_0).

        """
        aeqd_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {
                "name": "Modified Azimuthal Equidistant",
                "id": {"authority": "EPSG", "code": 9832},
            },
            "parameters": [
                {
                    "name": "Latitude of natural origin",
                    "value": latitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8801},
                },
                {
                    "name": "Longitude of natural origin",
                    "value": longitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8802},
                },
                {
                    "name": "False easting",
                    "value": false_easting,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8806},
                },
                {
                    "name": "False northing",
                    "value": false_northing,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8807},
                },
            ],
        }
        return cls.from_json_dict(aeqd_json)


class GeostationarySatelliteConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Geostationary Satellite conversion.

    :ref:`PROJ docs <geos>`
    """

    def __new__(
        cls,
        sweep_angle_axis: str,
        satellite_height: float,
        latitude_natural_origin: float = 0.0,
        longitude_natural_origin: float = 0.0,
        false_easting: float = 0.0,
        false_northing: float = 0.0,
    ):
        """
        Parameters
        ----------
        sweep_angle_axis: str
            Sweep angle axis of the viewing instrument. Valid options are “X” and “Y”.
        satellite_height: float
            Satellite height.
        latitude_natural_origin: float, default=0.0
            Latitude of projection center (lat_0).
        longitude_natural_origin: float, default=0.0
            Longitude of projection center (lon_0).
        false_easting: float, default=0.0
            False easting (x_0).
        false_northing: float, default=0.0
            False northing (y_0).

        """
        sweep_angle_axis = sweep_angle_axis.strip().upper()
        valid_sweep_axis = ("X", "Y")
        if sweep_angle_axis not in valid_sweep_axis:
            raise CRSError(f"sweep_angle_axis only supports {valid_sweep_axis}")

        if latitude_natural_origin != 0:
            warnings.warn(
                "The latitude of natural origin (lat_0) is not used "
                "within PROJ. It is only supported for exporting to "
                "the WKT or PROJ JSON formats."
            )

        geos_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {"name": f"Geostationary Satellite (Sweep {sweep_angle_axis})"},
            "parameters": [
                {
                    "name": "Satellite height",
                    "value": satellite_height,
                    "unit": "metre",
                },
                {
                    "name": "Latitude of natural origin",
                    "value": latitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8801},
                },
                {
                    "name": "Longitude of natural origin",
                    "value": longitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8802},
                },
                {
                    "name": "False easting",
                    "value": false_easting,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8806},
                },
                {
                    "name": "False northing",
                    "value": false_northing,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8807},
                },
            ],
        }
        return cls.from_json_dict(geos_json)


class LambertAzimuthalEqualAreaConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0 LambertAzumuthalEqualAreaConversion
    .. versionadded:: 3.2.0 LambertAzimuthalEqualAreaConversion

    Class for constructing the Lambert Azimuthal Equal Area conversion.

    :ref:`PROJ docs <laea>`
    """

    def __new__(
        cls,
        latitude_natural_origin: float = 0.0,
        longitude_natural_origin: float = 0.0,
        false_easting: float = 0.0,
        false_northing: float = 0.0,
    ):
        """
        Parameters
        ----------
        latitude_natural_origin: float, default=0.0
            Latitude of projection center (lat_0).
        longitude_natural_origin: float, default=0.0
            Longitude of projection center (lon_0).
        false_easting: float, default=0.0
            False easting (x_0).
        false_northing: float, default=0.0
            False northing (y_0).

        """
        laea_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {
                "name": "Lambert Azimuthal Equal Area",
                "id": {"authority": "EPSG", "code": 9820},
            },
            "parameters": [
                {
                    "name": "Latitude of natural origin",
                    "value": latitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8801},
                },
                {
                    "name": "Longitude of natural origin",
                    "value": longitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8802},
                },
                {
                    "name": "False easting",
                    "value": false_easting,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8806},
                },
                {
                    "name": "False northing",
                    "value": false_northing,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8807},
                },
            ],
        }
        return cls.from_json_dict(laea_json)


class LambertConformalConic2SPConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Lambert Conformal Conic 2SP conversion.

    :ref:`PROJ docs <lcc>`
    """

    def __new__(
        cls,
        latitude_first_parallel: float,
        latitude_second_parallel: float,
        latitude_false_origin: float = 0.0,
        longitude_false_origin: float = 0.0,
        easting_false_origin: float = 0.0,
        northing_false_origin: float = 0.0,
    ):
        """
        Parameters
        ----------
        latitude_first_parallel: float
            Latitude of 1st standard parallel (lat_1).
        latitude_second_parallel: float
            Latitude of 2nd standard parallel (lat_2).
        latitude_false_origin: float, default=0.0
            Latitude of projection center (lat_0).
        longitude_false_origin: float, default=0.0
            Longitude of projection center (lon_0).
        easting_false_origin: float, default=0.0
            False easting (x_0).
        northing_false_origin: float, default=0.0
            False northing (y_0).

        """
        lcc_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {
                "name": "Lambert Conic Conformal (2SP)",
                "id": {"authority": "EPSG", "code": 9802},
            },
            "parameters": [
                {
                    "name": "Latitude of 1st standard parallel",
                    "value": latitude_first_parallel,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8823},
                },
                {
                    "name": "Latitude of 2nd standard parallel",
                    "value": latitude_second_parallel,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8824},
                },
                {
                    "name": "Latitude of false origin",
                    "value": latitude_false_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8821},
                },
                {
                    "name": "Longitude of false origin",
                    "value": longitude_false_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8822},
                },
                {
                    "name": "Easting at false origin",
                    "value": easting_false_origin,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8826},
                },
                {
                    "name": "Northing at false origin",
                    "value": northing_false_origin,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8827},
                },
            ],
        }
        return cls.from_json_dict(lcc_json)


class LambertConformalConic1SPConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Lambert Conformal Conic 1SP conversion.

    :ref:`PROJ docs <lcc>`
    """

    def __new__(
        cls,
        latitude_natural_origin: float = 0.0,
        longitude_natural_origin: float = 0.0,
        false_easting: float = 0.0,
        false_northing: float = 0.0,
        scale_factor_natural_origin: float = 1.0,
    ):
        """
        Parameters
        ----------
        latitude_natural_origin: float, default=0.0
            Latitude of projection center (lat_0).
        longitude_natural_origin: float, default=0.0
            Longitude of projection center (lon_0).
        false_easting: float, default=0.0
            False easting (x_0).
        false_northing: float, default=0.0
            False northing (y_0).
        scale_factor_natural_origin: float, default=1.0
            Scale factor at natural origin (k_0).

        """
        lcc_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {
                "name": "Lambert Conic Conformal (1SP)",
                "id": {"authority": "EPSG", "code": 9801},
            },
            "parameters": [
                {
                    "name": "Latitude of natural origin",
                    "value": latitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8801},
                },
                {
                    "name": "Longitude of natural origin",
                    "value": longitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8802},
                },
                {
                    "name": "Scale factor at natural origin",
                    "value": scale_factor_natural_origin,
                    "unit": "unity",
                    "id": {"authority": "EPSG", "code": 8805},
                },
                {
                    "name": "False easting",
                    "value": false_easting,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8806},
                },
                {
                    "name": "False northing",
                    "value": false_northing,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8807},
                },
            ],
        }
        return cls.from_json_dict(lcc_json)


class LambertCylindricalEqualAreaConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Lambert Cylindrical Equal Area conversion.

    :ref:`PROJ docs <cea>`
    """

    def __new__(
        cls,
        latitude_first_parallel: float = 0.0,
        longitude_natural_origin: float = 0.0,
        false_easting: float = 0.0,
        false_northing: float = 0.0,
    ):
        """
        Parameters
        ----------
        latitude_first_parallel: float, default=0.0
            Latitude of 1st standard parallel (lat_ts).
        longitude_natural_origin: float, default=0.0
            Longitude of projection center (lon_0).
        false_easting: float, default=0.0
            False easting (x_0).
        false_northing: float, default=0.0
            False northing (y_0).

        """
        cea_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {
                "name": "Lambert Cylindrical Equal Area",
                "id": {"authority": "EPSG", "code": 9835},
            },
            "parameters": [
                {
                    "name": "Latitude of 1st standard parallel",
                    "value": latitude_first_parallel,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8823},
                },
                {
                    "name": "Longitude of natural origin",
                    "value": longitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8802},
                },
                {
                    "name": "False easting",
                    "value": false_easting,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8806},
                },
                {
                    "name": "False northing",
                    "value": false_northing,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8807},
                },
            ],
        }
        return cls.from_json_dict(cea_json)


class LambertCylindricalEqualAreaScaleConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Lambert Cylindrical Equal Area conversion.

    This version uses the scale factor and differs from the official version.

    The scale factor will be converted to the Latitude of 1st standard parallel (lat_ts)
    when exporting to WKT in PROJ>=7.0.0. Previous version will export it as a
    PROJ-based coordinate operation in the WKT.

    :ref:`PROJ docs <cea>`
    """

    def __new__(
        cls,
        longitude_natural_origin: float = 0.0,
        false_easting: float = 0.0,
        false_northing: float = 0.0,
        scale_factor_natural_origin: float = 1.0,
    ):
        """
        Parameters
        ----------
        longitude_natural_origin: float, default=0.0
            Longitude of projection center (lon_0).
        false_easting: float, default=0.0
            False easting (x_0).
        false_northing: float, default=0.0
            False northing (y_0).
        scale_factor_natural_origin: float, default=1.0
            Scale factor at natural origin (k or k_0).

        """
        # pylint: disable=import-outside-toplevel
        from pyproj.crs import CRS

        # hack due to: https://github.com/OSGeo/PROJ/issues/1881
        proj_string = (
            "+proj=cea "
            f"+lon_0={longitude_natural_origin} "
            f"+x_0={false_easting} "
            f"+y_0={false_northing} "
            f"+k_0={scale_factor_natural_origin}"
        )
        return cls.from_json(
            CRS(proj_string).coordinate_operation.to_json()  # type: ignore
        )


class MercatorAConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Mercator (variant A) conversion.

    :ref:`PROJ docs <merc>`
    """

    def __new__(
        cls,
        latitude_natural_origin: float = 0.0,
        longitude_natural_origin: float = 0.0,
        false_easting: float = 0.0,
        false_northing: float = 0.0,
        scale_factor_natural_origin: float = 1.0,
    ):
        """
        Parameters
        ----------
        latitude_natural_origin: float, default=0.0
            Latitude of natural origin (lat_0). Must be 0 by `this conversion's
            definition
            <https://epsg.org/coord-operation-method_9804/Mercator-variant-A.html>`_.
        longitude_natural_origin: float, default=0.0
            Longitude of natural origin (lon_0).
        false_easting: float, default=0.0
            False easting (x_0).
        false_northing: float, default=0.0
            False northing (y_0).
        scale_factor_natural_origin: float, default=1.0
            Scale factor at natural origin (k or k_0).

        """
        if latitude_natural_origin != 0:
            raise CRSError(
                "This conversion is defined for only latitude_natural_origin = 0."
            )
        merc_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {
                "name": "Mercator (variant A)",
                "id": {"authority": "EPSG", "code": 9804},
            },
            "parameters": [
                {
                    "name": "Latitude of natural origin",
                    "value": latitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8801},
                },
                {
                    "name": "Longitude of natural origin",
                    "value": longitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8802},
                },
                {
                    "name": "Scale factor at natural origin",
                    "value": scale_factor_natural_origin,
                    "unit": "unity",
                    "id": {"authority": "EPSG", "code": 8805},
                },
                {
                    "name": "False easting",
                    "value": false_easting,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8806},
                },
                {
                    "name": "False northing",
                    "value": false_northing,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8807},
                },
            ],
        }
        return cls.from_json_dict(merc_json)


class MercatorBConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Mercator (variant B) conversion.

    :ref:`PROJ docs <merc>`
    """

    def __new__(
        cls,
        latitude_first_parallel: float = 0.0,
        longitude_natural_origin: float = 0.0,
        false_easting: float = 0.0,
        false_northing: float = 0.0,
    ):
        """
        Parameters
        ----------
        latitude_first_parallel: float, default=0.0
            Latitude of 1st standard parallel (lat_ts).
        longitude_natural_origin: float, default=0.0
            Longitude of projection center (lon_0).
        false_easting: float, default=0.0
            False easting (x_0).
        false_northing: float, default=0.0
            False northing (y_0).

        """
        merc_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {
                "name": "Mercator (variant B)",
                "id": {"authority": "EPSG", "code": 9805},
            },
            "parameters": [
                {
                    "name": "Latitude of 1st standard parallel",
                    "value": latitude_first_parallel,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8823},
                },
                {
                    "name": "Longitude of natural origin",
                    "value": longitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8802},
                },
                {
                    "name": "False easting",
                    "value": false_easting,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8806},
                },
                {
                    "name": "False northing",
                    "value": false_northing,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8807},
                },
            ],
        }
        return cls.from_json_dict(merc_json)


class HotineObliqueMercatorBConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Hotine Oblique Mercator (variant B) conversion.

    :ref:`PROJ docs <omerc>`
    """

    def __new__(
        cls,
        latitude_projection_centre: float,
        longitude_projection_centre: float,
        azimuth_initial_line: float,
        angle_from_rectified_to_skew_grid: float,
        scale_factor_on_initial_line: float = 1.0,
        easting_projection_centre: float = 0.0,
        northing_projection_centre: float = 0.0,
    ):
        """
        Parameters
        ----------
        latitude_projection_centre: float
            Latitude of projection centre (lat_0).
        longitude_projection_centre: float
            Longitude of projection centre (lonc).
        azimuth_initial_line: float
            Azimuth of initial line (alpha).
        angle_from_rectified_to_skew_grid: float
            Angle from Rectified to Skew Grid (gamma).
        scale_factor_on_initial_line: float, default=1.0
            Scale factor on initial line (k or k_0).
        easting_projection_centre: float, default=0.0
            Easting at projection centre (x_0).
        northing_projection_centre: float, default=0.0
            Northing at projection centre (y_0).
        """
        omerc_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {
                "name": "Hotine Oblique Mercator (variant B)",
                "id": {"authority": "EPSG", "code": 9815},
            },
            "parameters": [
                {
                    "name": "Latitude of projection centre",
                    "value": latitude_projection_centre,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8811},
                },
                {
                    "name": "Longitude of projection centre",
                    "value": longitude_projection_centre,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8812},
                },
                {
                    "name": "Azimuth of initial line",
                    "value": azimuth_initial_line,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8813},
                },
                {
                    "name": "Angle from Rectified to Skew Grid",
                    "value": angle_from_rectified_to_skew_grid,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8814},
                },
                {
                    "name": "Scale factor on initial line",
                    "value": scale_factor_on_initial_line,
                    "unit": "unity",
                    "id": {"authority": "EPSG", "code": 8815},
                },
                {
                    "name": "Easting at projection centre",
                    "value": easting_projection_centre,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8816},
                },
                {
                    "name": "Northing at projection centre",
                    "value": northing_projection_centre,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8817},
                },
            ],
        }
        return cls.from_json_dict(omerc_json)


class OrthographicConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Orthographic conversion.

    :ref:`PROJ docs <ortho>`
    """

    def __new__(
        cls,
        latitude_natural_origin: float = 0.0,
        longitude_natural_origin: float = 0.0,
        false_easting: float = 0.0,
        false_northing: float = 0.0,
    ):
        """
        Parameters
        ----------
        latitude_natural_origin: float, default=0.0
            Latitude of projection center (lat_0).
        longitude_natural_origin: float, default=0.0
            Longitude of projection center (lon_0).
        false_easting: float, default=0.0
            False easting (x_0).
        false_northing: float, default=0.0
            False northing (y_0).

        """
        ortho_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {
                "name": "Orthographic",
                "id": {"authority": "EPSG", "code": 9840},
            },
            "parameters": [
                {
                    "name": "Latitude of natural origin",
                    "value": latitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8801},
                },
                {
                    "name": "Longitude of natural origin",
                    "value": longitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8802},
                },
                {
                    "name": "False easting",
                    "value": false_easting,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8806},
                },
                {
                    "name": "False northing",
                    "value": false_northing,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8807},
                },
            ],
        }
        return cls.from_json_dict(ortho_json)


class PolarStereographicAConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Polar Stereographic A conversion.

    :ref:`PROJ docs <stere>`
    """

    def __new__(
        cls,
        latitude_natural_origin: float,
        longitude_natural_origin: float = 0.0,
        false_easting: float = 0.0,
        false_northing: float = 0.0,
        scale_factor_natural_origin: float = 1.0,
    ):
        """
        Parameters
        ----------
        latitude_natural_origin: float
            Latitude of natural origin (lat_0). Either +90 or -90.
        longitude_natural_origin: float, default=0.0
            Longitude of natural origin (lon_0).
        false_easting: float, default=0.0
            False easting (x_0).
        false_northing: float, default=0.0
            False northing (y_0).
        scale_factor_natural_origin: float, default=0.0
            Scale factor at natural origin (k or k_0).

        """

        stere_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {
                "name": "Polar Stereographic (variant A)",
                "id": {"authority": "EPSG", "code": 9810},
            },
            "parameters": [
                {
                    "name": "Latitude of natural origin",
                    "value": latitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8801},
                },
                {
                    "name": "Longitude of natural origin",
                    "value": longitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8802},
                },
                {
                    "name": "Scale factor at natural origin",
                    "value": scale_factor_natural_origin,
                    "unit": "unity",
                    "id": {"authority": "EPSG", "code": 8805},
                },
                {
                    "name": "False easting",
                    "value": false_easting,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8806},
                },
                {
                    "name": "False northing",
                    "value": false_northing,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8807},
                },
            ],
        }
        return cls.from_json_dict(stere_json)


class PolarStereographicBConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Polar Stereographic B conversion.

    :ref:`PROJ docs <stere>`
    """

    def __new__(
        cls,
        latitude_standard_parallel: float = 0.0,
        longitude_origin: float = 0.0,
        false_easting: float = 0.0,
        false_northing: float = 0.0,
    ):
        """
        Parameters
        ----------
        latitude_standard_parallel: float, default=0.0
            Latitude of standard parallel (lat_ts).
        longitude_origin: float, default=0.0
            Longitude of origin (lon_0).
        false_easting: float, default=0.0
            False easting (x_0).
        false_northing: float, default=0.0
            False northing (y_0).

        """
        stere_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {
                "name": "Polar Stereographic (variant B)",
                "id": {"authority": "EPSG", "code": 9829},
            },
            "parameters": [
                {
                    "name": "Latitude of standard parallel",
                    "value": latitude_standard_parallel,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8832},
                },
                {
                    "name": "Longitude of origin",
                    "value": longitude_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8833},
                },
                {
                    "name": "False easting",
                    "value": false_easting,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8806},
                },
                {
                    "name": "False northing",
                    "value": false_northing,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8807},
                },
            ],
        }
        return cls.from_json_dict(stere_json)


class SinusoidalConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Sinusoidal conversion.

    :ref:`PROJ docs <sinu>`
    """

    def __new__(
        cls,
        longitude_natural_origin: float = 0.0,
        false_easting: float = 0.0,
        false_northing: float = 0.0,
    ):
        """
        Parameters
        ----------
        longitude_natural_origin: float, default=0.0
            Longitude of projection center (lon_0).
        false_easting: float, default=0.0
            False easting (x_0).
        false_northing: float, default=0.0
            False northing (y_0).

        """
        sinu_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {"name": "Sinusoidal"},
            "parameters": [
                {
                    "name": "Longitude of natural origin",
                    "value": longitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8802},
                },
                {
                    "name": "False easting",
                    "value": false_easting,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8806},
                },
                {
                    "name": "False northing",
                    "value": false_northing,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8807},
                },
            ],
        }
        return cls.from_json_dict(sinu_json)


class StereographicConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Stereographic conversion.

    :ref:`PROJ docs <stere>`
    """

    def __new__(
        cls,
        latitude_natural_origin: float = 0.0,
        longitude_natural_origin: float = 0.0,
        false_easting: float = 0.0,
        false_northing: float = 0.0,
        scale_factor_natural_origin: float = 1.0,
    ):
        """
        Parameters
        ----------
        latitude_natural_origin: float, default=0.0
            Latitude of natural origin (lat_0).
        longitude_natural_origin: float, default=0.0
            Longitude of natural origin (lon_0).
        false_easting: float, default=0.0
            False easting (x_0).
        false_northing: float, default=0.0
            False northing (y_0).
        scale_factor_natural_origin: float, default=1.0
            Scale factor at natural origin (k or k_0).

        """

        stere_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {"name": "Stereographic"},
            "parameters": [
                {
                    "name": "Latitude of natural origin",
                    "value": latitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8801},
                },
                {
                    "name": "Longitude of natural origin",
                    "value": longitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8802},
                },
                {
                    "name": "Scale factor at natural origin",
                    "value": scale_factor_natural_origin,
                    "unit": "unity",
                    "id": {"authority": "EPSG", "code": 8805},
                },
                {
                    "name": "False easting",
                    "value": false_easting,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8806},
                },
                {
                    "name": "False northing",
                    "value": false_northing,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8807},
                },
            ],
        }
        return cls.from_json_dict(stere_json)


class UTMConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the UTM conversion.

    :ref:`PROJ docs <utm>`
    """

    def __new__(cls, zone: str, hemisphere: str = "N"):
        """
        Parameters
        ----------
        zone: int
            UTM Zone between 1-60.
        hemisphere: str, default="N"
            Either N for North or S for South.
        """
        return cls.from_name(f"UTM zone {zone}{hemisphere}")


class TransverseMercatorConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Transverse Mercator conversion.

    :ref:`PROJ docs <tmerc>`
    """

    def __new__(
        cls,
        latitude_natural_origin: float = 0.0,
        longitude_natural_origin: float = 0.0,
        false_easting: float = 0.0,
        false_northing: float = 0.0,
        scale_factor_natural_origin: float = 1.0,
    ):
        """
        Parameters
        ----------
        latitude_natural_origin: float, default=0.0
            Latitude of projection center (lat_0).
        longitude_natural_origin: float, default=0.0
            Longitude of projection center (lon_0).
        false_easting: float, default=0.0
            False easting (x_0).
        false_northing: float, default=0.0
            False northing (y_0).
        scale_factor_natural_origin: float, default=1.0
            Scale factor at natural origin (k or k_0).

        """
        tmerc_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {
                "name": "Transverse Mercator",
                "id": {"authority": "EPSG", "code": 9807},
            },
            "parameters": [
                {
                    "name": "Latitude of natural origin",
                    "value": latitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8801},
                },
                {
                    "name": "Longitude of natural origin",
                    "value": longitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8802},
                },
                {
                    "name": "Scale factor at natural origin",
                    "value": scale_factor_natural_origin,
                    "unit": "unity",
                    "id": {"authority": "EPSG", "code": 8805},
                },
                {
                    "name": "False easting",
                    "value": false_easting,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8806},
                },
                {
                    "name": "False northing",
                    "value": false_northing,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8807},
                },
            ],
        }
        return cls.from_json_dict(tmerc_json)


class VerticalPerspectiveConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Vetical Perspective conversion.

    :ref:`PROJ docs <nsper>`
    """

    def __new__(
        cls,
        viewpoint_height: float,
        latitude_topocentric_origin: float = 0.0,
        longitude_topocentric_origin: float = 0.0,
        ellipsoidal_height_topocentric_origin: float = 0.0,
        false_easting: float = 0.0,
        false_northing: float = 0.0,
    ):
        """
        Parameters
        ----------
        viewpoint_height: float
            Viewpoint height (h).
        latitude_topocentric_origin: float, default=0.0
            Latitude of topocentric origin (lat_0).
        longitude_topocentric_origin: float, default=0.0
            Longitude of topocentric origin (lon_0).
        ellipsoidal_height_topocentric_origin: float, default=0.0
            Ellipsoidal height of topocentric origin.
        false_easting: float, default=0.0
            False easting (x_0).
        false_northing: float, default=0.0
            False northing (y_0).

        """
        nsper_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {
                "name": "Vertical Perspective",
                "id": {"authority": "EPSG", "code": 9838},
            },
            "parameters": [
                {
                    "name": "Latitude of topocentric origin",
                    "value": latitude_topocentric_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8834},
                },
                {
                    "name": "Longitude of topocentric origin",
                    "value": longitude_topocentric_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8835},
                },
                {
                    "name": "Ellipsoidal height of topocentric origin",
                    "value": ellipsoidal_height_topocentric_origin,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8836},
                },
                {
                    "name": "Viewpoint height",
                    "value": viewpoint_height,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8840},
                },
                {
                    "name": "False easting",
                    "value": false_easting,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8806},
                },
                {
                    "name": "False northing",
                    "value": false_northing,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8807},
                },
            ],
        }
        return cls.from_json_dict(nsper_json)


class RotatedLatitudeLongitudeConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Rotated Latitude Longitude conversion.

    :ref:`PROJ docs <ob_tran>`
    """

    def __new__(cls, o_lat_p: float, o_lon_p: float, lon_0: float = 0.0):
        """
        Parameters
        ----------
        o_lat_p: float
            Latitude of the North pole of the unrotated source CRS,
            expressed in the rotated geographic CRS.
        o_lon_p: float
            Longitude of the North pole of the unrotated source CRS,
            expressed in the rotated geographic CRS.
        lon_0: float, default=0.0
            Longitude of projection center.

        """
        rot_latlon_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {"name": "PROJ ob_tran o_proj=longlat"},
            "parameters": [
                {"name": "o_lat_p", "value": o_lat_p, "unit": "degree"},
                {"name": "o_lon_p", "value": o_lon_p, "unit": "degree"},
                {"name": "lon_0", "value": lon_0, "unit": "degree"},
            ],
        }
        return cls.from_json_dict(rot_latlon_json)


class PoleRotationNetCDFCFConversion(CoordinateOperation):
    """
    .. versionadded:: 3.3.0

    Class for constructing the Pole rotation (netCDF CF convention) conversion.

    http://cfconventions.org/cf-conventions/cf-conventions.html#_rotated_pole

    :ref:`PROJ docs <ob_tran>`
    """

    def __new__(
        cls,
        grid_north_pole_latitude: float,
        grid_north_pole_longitude: float,
        north_pole_grid_longitude: float = 0.0,
    ):
        """
        Parameters
        ----------
        grid_north_pole_latitude: float
            Latitude of the North pole of the unrotated source CRS,
            expressed in the rotated geographic CRS (o_lat_p)
        grid_north_pole_longitude: float
            Longitude of projection center (lon_0 - 180).
        north_pole_grid_longitude: float, default=0.0
            Longitude of the North pole of the unrotated source CRS,
            expressed in the rotated geographic CRS (o_lon_p).
        """
        rot_latlon_json = {
            "$schema": "https://proj.org/schemas/v0.4/projjson.schema.json",
            "type": "Conversion",
            "name": "Pole rotation (netCDF CF convention)",
            "method": {"name": "Pole rotation (netCDF CF convention)"},
            "parameters": [
                {
                    "name": "Grid north pole latitude (netCDF CF convention)",
                    "value": grid_north_pole_latitude,
                    "unit": "degree",
                },
                {
                    "name": "Grid north pole longitude (netCDF CF convention)",
                    "value": grid_north_pole_longitude,
                    "unit": "degree",
                },
                {
                    "name": "North pole grid longitude (netCDF CF convention)",
                    "value": north_pole_grid_longitude,
                    "unit": "degree",
                },
            ],
        }
        return cls.from_json_dict(rot_latlon_json)


class EquidistantCylindricalConversion(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the Equidistant Cylintrical (Plate Carrée) conversion.

    :ref:`PROJ docs <eqc>`
    """

    def __new__(
        cls,
        latitude_first_parallel: float = 0.0,
        latitude_natural_origin: float = 0.0,
        longitude_natural_origin: float = 0.0,
        false_easting: float = 0.0,
        false_northing: float = 0.0,
    ):
        """
        Parameters
        ----------
        latitude_first_parallel: float, default=0.0
            Latitude of 1st standard parallel (lat_ts).
        latitude_natural_origin: float, default=0.0
            Longitude of projection center (lon_0).
        longitude_natural_origin: float, default=0.0
            Longitude of projection center (lon_0).
        false_easting: float, default=0.0
            False easting (x_0).
        false_northing: float, default=0.0
            False northing (y_0).
        """
        eqc_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Conversion",
            "name": "unknown",
            "method": {
                "name": "Equidistant Cylindrical",
                "id": {"authority": "EPSG", "code": 1028},
            },
            "parameters": [
                {
                    "name": "Latitude of 1st standard parallel",
                    "value": latitude_first_parallel,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8823},
                },
                {
                    "name": "Latitude of natural origin",
                    "value": latitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8801},
                },
                {
                    "name": "Longitude of natural origin",
                    "value": longitude_natural_origin,
                    "unit": "degree",
                    "id": {"authority": "EPSG", "code": 8802},
                },
                {
                    "name": "False easting",
                    "value": false_easting,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8806},
                },
                {
                    "name": "False northing",
                    "value": false_northing,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8807},
                },
            ],
        }
        return cls.from_json_dict(eqc_json)


# Add an alias for PlateCarree
PlateCarreeConversion = EquidistantCylindricalConversion


class ToWGS84Transformation(CoordinateOperation):
    """
    .. versionadded:: 2.5.0

    Class for constructing the ToWGS84 Transformation.
    """

    def __new__(
        cls,
        source_crs: Any,
        x_axis_translation: float = 0,
        y_axis_translation: float = 0,
        z_axis_translation: float = 0,
        x_axis_rotation: float = 0,
        y_axis_rotation: float = 0,
        z_axis_rotation: float = 0,
        scale_difference: float = 0,
    ):
        """
        Parameters
        ----------
        source_crs: Any
            Input to create the Source CRS.
        x_axis_translation: float, default=0.0
            X-axis translation.
        y_axis_translation: float, default=0.0
            Y-axis translation.
        z_axis_translation: float, default=0.0
            Z-axis translation.
        x_axis_rotation: float, default=0.0
            X-axis rotation.
        y_axis_rotation: float, default=0.0
            Y-axis rotation.
        z_axis_rotation: float, default=0.0
            Z-axis rotation.
        scale_difference: float, default=0.0
            Scale difference.
        """
        # pylint: disable=import-outside-toplevel
        from pyproj.crs import CRS

        towgs84_json = {
            "$schema": "https://proj.org/schemas/v0.2/projjson.schema.json",
            "type": "Transformation",
            "name": "Transformation from unknown to WGS84",
            "source_crs": CRS.from_user_input(source_crs).to_json_dict(),
            "target_crs": {
                "type": "GeographicCRS",
                "name": "WGS 84",
                "datum": {
                    "type": "GeodeticReferenceFrame",
                    "name": "World Geodetic System 1984",
                    "ellipsoid": {
                        "name": "WGS 84",
                        "semi_major_axis": 6378137,
                        "inverse_flattening": 298.257223563,
                    },
                },
                "coordinate_system": {
                    "subtype": "ellipsoidal",
                    "axis": [
                        {
                            "name": "Latitude",
                            "abbreviation": "lat",
                            "direction": "north",
                            "unit": "degree",
                        },
                        {
                            "name": "Longitude",
                            "abbreviation": "lon",
                            "direction": "east",
                            "unit": "degree",
                        },
                    ],
                },
                "id": {"authority": "EPSG", "code": 4326},
            },
            "method": {
                "name": "Position Vector transformation (geog2D domain)",
                "id": {"authority": "EPSG", "code": 9606},
            },
            "parameters": [
                {
                    "name": "X-axis translation",
                    "value": x_axis_translation,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8605},
                },
                {
                    "name": "Y-axis translation",
                    "value": y_axis_translation,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8606},
                },
                {
                    "name": "Z-axis translation",
                    "value": z_axis_translation,
                    "unit": "metre",
                    "id": {"authority": "EPSG", "code": 8607},
                },
                {
                    "name": "X-axis rotation",
                    "value": x_axis_rotation,
                    "unit": {
                        "type": "AngularUnit",
                        "name": "arc-second",
                        "conversion_factor": 4.84813681109536e-06,
                    },
                    "id": {"authority": "EPSG", "code": 8608},
                },
                {
                    "name": "Y-axis rotation",
                    "value": y_axis_rotation,
                    "unit": {
                        "type": "AngularUnit",
                        "name": "arc-second",
                        "conversion_factor": 4.84813681109536e-06,
                    },
                    "id": {"authority": "EPSG", "code": 8609},
                },
                {
                    "name": "Z-axis rotation",
                    "value": z_axis_rotation,
                    "unit": {
                        "type": "AngularUnit",
                        "name": "arc-second",
                        "conversion_factor": 4.84813681109536e-06,
                    },
                    "id": {"authority": "EPSG", "code": 8610},
                },
                {
                    "name": "Scale difference",
                    "value": scale_difference,
                    "unit": {
                        "type": "ScaleUnit",
                        "name": "parts per million",
                        "conversion_factor": 1e-06,
                    },
                    "id": {"authority": "EPSG", "code": 8611},
                },
            ],
        }

        return cls.from_json_dict(towgs84_json)



include "proj.pxi"


def get_proj_operations_map():
    """
    Returns
    -------
    dict:
        Operations supported by PROJ.
    """
    cdef const PJ_OPERATIONS *proj_operations = proj_list_operations()
    cdef int iii = 0
    operations_map = {}
    while proj_operations[iii].id != NULL:
        operations_map[proj_operations[iii].id] = \
            proj_operations[iii].descr[0].split("\n\t")[0]
        iii += 1
    return operations_map


def get_ellps_map():
    """
    Returns
    -------
    dict:
        Ellipsoids supported by PROJ.
    """
    cdef const PJ_ELLPS *proj_ellps = proj_list_ellps()
    cdef int iii = 0
    ellps_map = {}
    while proj_ellps[iii].id != NULL:
        major_key, major_val = proj_ellps[iii].major.split("=")
        ell_key, ell_val = proj_ellps[iii].ell.split("=")
        ellps_map[proj_ellps[iii].id] = {
            major_key: float(major_val),
            ell_key: float(ell_val),
            "description": proj_ellps[iii].name
        }
        iii += 1
    return ellps_map


def get_prime_meridians_map():
    """
    Returns
    -------
    dict:
        Prime Meridians supported by PROJ.
    """
    cdef const PJ_PRIME_MERIDIANS *prime_meridians = proj_list_prime_meridians()
    cdef int iii = 0
    prime_meridians_map = {}
    while prime_meridians[iii].id != NULL:
        prime_meridians_map[prime_meridians[iii].id] = \
            prime_meridians[iii].defn
        iii += 1
    return prime_meridians_map



"""
Based on the logic in the PROJ projsync CLI program

https://github.com/OSGeo/PROJ/blob/9ff543c4ffd86152bc58d0a0164b2ce9ebbb8bec/src/apps/projsync.cpp
"""
import hashlib
import json
import os
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import Any, Optional
from urllib.request import urlretrieve

from pyproj._sync import get_proj_endpoint
from pyproj.aoi import BBox
from pyproj.datadir import get_data_dir, get_user_data_dir


def _bbox_from_coords(coords: list) -> Optional[BBox]:
    """
    Get the bounding box from coordinates
    """
    try:
        xxx, yyy = zip(*coords)
        return BBox(west=min(xxx), south=min(yyy), east=max(xxx), north=max(yyy))
    except ValueError:
        pass
    coord_bbox = None
    for coord_set in coords:
        bbox = _bbox_from_coords(coord_set)
        if coord_bbox is None:
            coord_bbox = bbox
        else:
            coord_bbox.west = min(coord_bbox.west, bbox.west)
            coord_bbox.south = min(coord_bbox.south, bbox.south)
            coord_bbox.north = max(coord_bbox.north, bbox.north)
            coord_bbox.east = max(coord_bbox.east, bbox.east)
    return coord_bbox


def _bbox_from_geom(geom: dict[str, Any]) -> Optional[BBox]:
    """
    Get the bounding box from geojson geometry
    """
    if "coordinates" not in geom or "type" not in geom:
        return None
    coordinates = geom["coordinates"]
    if geom["type"] != "MultiPolygon":
        return _bbox_from_coords(coordinates)
    found_minus_180 = False
    found_plus_180 = False
    bboxes = []
    for coordinate_set in coordinates:
        bbox = _bbox_from_coords(coordinate_set)
        if bbox is None:
            continue
        if bbox.west == -180:
            found_minus_180 = True
        elif bbox.east == 180:
            found_plus_180 = True
        bboxes.append(bbox)
    grid_bbox = None
    for bbox in bboxes:
        if found_minus_180 and found_plus_180 and bbox.west == -180:
            bbox.west = 180
            bbox.east += 360
        if grid_bbox is None:
            grid_bbox = bbox
        else:
            grid_bbox.west = min(grid_bbox.west, bbox.west)
            grid_bbox.south = min(grid_bbox.south, bbox.south)
            grid_bbox.north = max(grid_bbox.north, bbox.north)
            grid_bbox.east = max(grid_bbox.east, bbox.east)
    return grid_bbox


def _filter_bbox(
    feature: dict[str, Any], bbox: BBox, spatial_test: str, include_world_coverage: bool
) -> bool:
    """
    Filter by the bounding box. Designed to use with 'filter'
    """
    geom = feature.get("geometry")
    if geom is not None:
        geom_bbox = _bbox_from_geom(geom)
        if geom_bbox is None:
            return False
        if (
            geom_bbox.east - geom_bbox.west > 359
            and geom_bbox.north - geom_bbox.south > 179
        ):
            if not include_world_coverage:
                return False
            geom_bbox.west = -float("inf")
            geom_bbox.east = float("inf")
        elif geom_bbox.east > 180 and bbox.west < -180:
            geom_bbox.west -= 360
            geom_bbox.east -= 360
        return getattr(bbox, spatial_test)(geom_bbox)
    return False


def _filter_properties(
    feature: dict[str, Any],
    source_id: Optional[str] = None,
    area_of_use: Optional[str] = None,
    filename: Optional[str] = None,
) -> bool:
    """
    Filter by the properties. Designed to use with 'filter'
    """
    properties = feature.get("properties")
    if not properties:
        return False
    p_filename = properties.get("name")
    p_source_id = properties.get("source_id")
    if not p_filename or not p_source_id:
        return False
    source_id__matched = source_id is None or source_id in p_source_id
    area_of_use__matched = area_of_use is None or area_of_use in properties.get(
        "area_of_use", ""
    )
    filename__matched = filename is None or filename in p_filename
    if source_id__matched and area_of_use__matched and filename__matched:
        return True
    return False


def _is_download_needed(grid_name: str) -> bool:
    """
    Run through all of the PROJ directories to see if the
    file already exists.
    """
    if Path(get_user_data_dir(), grid_name).exists():
        return False
    for data_dir in get_data_dir().split(os.pathsep):
        if Path(data_dir, grid_name).exists():
            return False
    return True


def _filter_download_needed(feature: dict[str, Any]) -> bool:
    """
    Filter grids so only those that need to be downloaded are included.
    """
    properties = feature.get("properties")
    if not properties:
        return False
    filename = properties.get("name")
    if not filename:
        return False
    return _is_download_needed(filename)


def _sha256sum(input_file):
    """
    Return sha256 checksum of file given by path.
    """

    hasher = hashlib.sha256()
    with open(input_file, "rb") as file:
        for chunk in iter(lambda: file.read(65536), b""):
            hasher.update(chunk)

    return hasher.hexdigest()


def _download_resource_file(
    file_url, short_name, directory, verbose=False, sha256=None
):
    """
    Download resource file from PROJ url
    """
    if verbose:
        print(f"Downloading: {file_url}")
    tmp_path = Path(directory, f"{short_name}.part")
    try:
        urlretrieve(file_url, tmp_path)
        if sha256 is not None and sha256 != _sha256sum(tmp_path):
            raise RuntimeError(f"SHA256 mismatch: {short_name}")
        tmp_path.replace(Path(directory, short_name))
    finally:
        try:
            os.remove(tmp_path)
        except FileNotFoundError:
            pass


def _load_grid_geojson(target_directory=None) -> dict[str, Any]:
    """
    Returns
    -------
    dict[str, Any]:
        The PROJ grid data list.
    """
    if target_directory is None:
        target_directory = get_user_data_dir(True)
    local_path = Path(target_directory, "files.geojson")
    if not local_path.exists() or (
        (datetime.utcnow() - datetime.fromtimestamp(local_path.stat().st_mtime)).days
        > 0
    ):
        _download_resource_file(
            file_url=f"{get_proj_endpoint()}/files.geojson",
            short_name="files.geojson",
            directory=target_directory,
        )
    return json.loads(local_path.read_text(encoding="utf-8"))


def get_transform_grid_list(
    source_id: Optional[str] = None,
    area_of_use: Optional[str] = None,
    filename: Optional[str] = None,
    bbox: Optional[BBox] = None,
    spatial_test: str = "intersects",
    include_world_coverage: bool = True,
    include_already_downloaded: bool = False,
    target_directory: Optional[str] = None,
) -> tuple:
    """
    Get a list of transform grids that can be downloaded.

    Parameters
    ----------
    source_id: str, optional
    area_of_use: str, optional
    filename: str, optional
    bbox: BBox, optional
    spatial_test: str, default="intersects"
        Can be "contains" or "intersects".
    include_world_coverage: bool, default=True
        If True, it will include grids with a global extent.
    include_already_downloaded: bool, default=False
        If True, it will list grids regardless of if they are downloaded.
    target_directory: Union[str, Path, None], optional
        The directory to download the geojson file to.
        Default is the user writable directory.

    Returns
    -------
    list[dict[str, Any]]:
        A list of geojson data of containing information about features
        that can be downloaded.
    """
    features = _load_grid_geojson(target_directory=target_directory)["features"]
    if bbox is not None:
        if bbox.west > 180 and bbox.east > bbox.west:
            bbox.west -= 360
            bbox.east -= 360
        elif bbox.west < -180 and bbox.east > bbox.west:
            bbox.west += 360
            bbox.east += 360
        elif abs(bbox.west) < 180 and abs(bbox.east) < 180 and bbox.east < bbox.west:
            bbox.east += 360
        features = filter(
            partial(
                _filter_bbox,
                bbox=bbox,
                spatial_test=spatial_test,
                include_world_coverage=include_world_coverage,
            ),
            features,
        )
    # filter by properties
    features = filter(
        partial(
            _filter_properties,
            source_id=source_id,
            area_of_use=area_of_use,
            filename=filename,
        ),
        features,
    )
    if include_already_downloaded:
        return tuple(features)
    return tuple(filter(_filter_download_needed, features))



from cpython.ref cimport PyObject

from math import degrees, radians


cdef double _DG2RAD = radians(1.)
cdef double _RAD2DG = degrees(1.)
cdef int _DOUBLESIZE = sizeof(double)


cdef extern from "math.h" nogil:
    ctypedef enum:
        HUGE_VAL


cdef extern from "Python.h":
    ctypedef enum:
        PyBUF_WRITABLE
    int PyObject_GetBuffer(PyObject *exporter, Py_buffer *view, int flags)
    void PyBuffer_Release(Py_buffer *view)


cdef class PyBuffWriteManager:
    cdef Py_buffer buffer
    cdef double* data
    cdef public Py_ssize_t len

    def __cinit__(self):
        self.data = NULL

    def __init__(self, object data):
        if PyObject_GetBuffer(<PyObject *>data, &self.buffer, PyBUF_WRITABLE) <> 0:
            raise BufferError("pyproj had a problem getting the buffer from data.")
        self.data = <double *>self.buffer.buf
        self.len = self.buffer.len // self.buffer.itemsize

    def __dealloc__(self):
        PyBuffer_Release(&self.buffer)
        self.data = NULL



import pytest

from pyproj.crs.coordinate_system import (
    Cartesian2DCS,
    Ellipsoidal2DCS,
    Ellipsoidal3DCS,
    VerticalCS,
)
from pyproj.crs.enums import (
    Cartesian2DCSAxis,
    Ellipsoidal2DCSAxis,
    Ellipsoidal3DCSAxis,
    VerticalCSAxis,
)


@pytest.mark.parametrize(
    "axis, direction, unit_name",
    [
        ("UP", "up", "metre"),
        (VerticalCSAxis.UP, "up", "metre"),
        (VerticalCSAxis.UP_US_FT, "up", "US survey foot"),
        ("UP_FT", "up", "foot"),
        (VerticalCSAxis.DEPTH, "down", "metre"),
        (VerticalCSAxis.DEPTH_US_FT, "down", "US survey foot"),
        ("DEPTH_FT", "down", "foot"),
        (VerticalCSAxis.GRAVITY_HEIGHT_US_FT, "up", "US survey foot"),
        ("GRAVITY_HEIGHT_FT", "up", "foot"),
    ],
)
def test_vertical_cs(axis, direction, unit_name):
    vcs = VerticalCS(axis=axis)
    assert len(vcs.axis_list) == 1
    assert vcs.axis_list[0].direction == direction
    assert vcs.axis_list[0].unit_name == unit_name


@pytest.mark.parametrize(
    "axis, name_0, direction_0, name_1, direction_1, unit",
    [
        ("EASTING_NORTHING", "Easting", "east", "Northing", "north", "metre"),
        (
            Cartesian2DCSAxis.NORTHING_EASTING,
            "Northing",
            "north",
            "Easting",
            "east",
            "metre",
        ),
        ("EASTING_NORTHING_FT", "Easting", "east", "Northing", "north", "foot"),
        (
            Cartesian2DCSAxis.NORTHING_EASTING_FT,
            "Northing",
            "north",
            "Easting",
            "east",
            "foot",
        ),
        (
            "EASTING_NORTHING_US_FT",
            "Easting",
            "east",
            "Northing",
            "north",
            "US survey foot",
        ),
        (
            Cartesian2DCSAxis.NORTHING_EASTING_US_FT,
            "Northing",
            "north",
            "Easting",
            "east",
            "US survey foot",
        ),
        (
            "NORTH_POLE_EASTING_SOUTH_NORTHING_SOUTH",
            "Easting",
            "south",
            "Northing",
            "south",
            "metre",
        ),
        (
            Cartesian2DCSAxis.SOUTH_POLE_EASTING_NORTH_NORTHING_NORTH,
            "Easting",
            "north",
            "Northing",
            "north",
            "metre",
        ),
        ("WESTING_SOUTHING", "Easting", "west", "Northing", "south", "metre"),
    ],
)
def test_cartesian_2d_cs(axis, name_0, direction_0, name_1, direction_1, unit):
    vcs = Cartesian2DCS(axis=axis)
    assert len(vcs.axis_list) == 2
    assert vcs.axis_list[0].direction == direction_0
    assert vcs.axis_list[0].name == name_0
    assert vcs.axis_list[0].unit_name == unit
    assert vcs.axis_list[1].direction == direction_1
    assert vcs.axis_list[1].name == name_1
    assert vcs.axis_list[1].unit_name == unit


@pytest.mark.parametrize(
    "axis, name_0, direction_0, name_1, direction_1",
    [
        (
            Ellipsoidal2DCSAxis.LONGITUDE_LATITUDE,
            "Longitude",
            "east",
            "Latitude",
            "north",
        ),
        (
            Ellipsoidal2DCSAxis.LATITUDE_LONGITUDE,
            "Latitude",
            "north",
            "Longitude",
            "east",
        ),
    ],
)
def test_ellipsoidal_2d_cs(axis, name_0, direction_0, name_1, direction_1):
    vcs = Ellipsoidal2DCS(axis=axis)
    assert len(vcs.axis_list) == 2
    assert vcs.axis_list[0].direction == direction_0
    assert vcs.axis_list[0].name == name_0
    assert vcs.axis_list[0].unit_name == "degree"
    assert vcs.axis_list[1].direction == direction_1
    assert vcs.axis_list[1].name == name_1
    assert vcs.axis_list[1].unit_name == "degree"


@pytest.mark.parametrize(
    "axis, name_0, direction_0, name_1, direction_1",
    [
        (
            Ellipsoidal3DCSAxis.LONGITUDE_LATITUDE_HEIGHT,
            "Longitude",
            "east",
            "Latitude",
            "north",
        ),
        (
            Ellipsoidal3DCSAxis.LATITUDE_LONGITUDE_HEIGHT,
            "Latitude",
            "north",
            "Longitude",
            "east",
        ),
    ],
)
def test_ellipsoidal_3d_cs(axis, name_0, direction_0, name_1, direction_1):
    vcs = Ellipsoidal3DCS(axis=axis)
    assert len(vcs.axis_list) == 3
    assert vcs.axis_list[0].direction == direction_0
    assert vcs.axis_list[0].name == name_0
    assert vcs.axis_list[0].unit_name == "degree"
    assert vcs.axis_list[1].direction == direction_1
    assert vcs.axis_list[1].name == name_1
    assert vcs.axis_list[1].unit_name == "degree"
    assert vcs.axis_list[2].direction == "up"
    assert vcs.axis_list[2].name == "Ellipsoidal height"
    assert vcs.axis_list[2].unit_name == "metre"


def test_ellipsoidal2dcs_to_cf():
    ecs = Ellipsoidal2DCS(axis=Ellipsoidal2DCSAxis.LATITUDE_LONGITUDE)
    assert ecs.to_cf() == [
        {
            "standard_name": "latitude",
            "long_name": "latitude coordinate",
            "units": "degrees_north",
            "axis": "Y",
        },
        {
            "standard_name": "longitude",
            "long_name": "longitude coordinate",
            "units": "degrees_east",
            "axis": "X",
        },
    ]


def test_ellipsoidal3dcs_to_cf():
    ecs = Ellipsoidal3DCS(axis=Ellipsoidal3DCSAxis.LONGITUDE_LATITUDE_HEIGHT)
    assert ecs.to_cf() == [
        {
            "standard_name": "longitude",
            "long_name": "longitude coordinate",
            "units": "degrees_east",
            "axis": "X",
        },
        {
            "standard_name": "latitude",
            "long_name": "latitude coordinate",
            "units": "degrees_north",
            "axis": "Y",
        },
        {
            "standard_name": "height_above_reference_ellipsoid",
            "long_name": "Ellipsoidal height",
            "units": "metre",
            "positive": "up",
            "axis": "Z",
        },
    ]


def test_cartesian2dcs_ft_to_cf():
    csft = Cartesian2DCS(axis=Cartesian2DCSAxis.NORTHING_EASTING_FT)
    csft.to_cf() == [
        {
            "axis": "Y",
            "long_name": "Northing",
            "standard_name": "projection_y_coordinate",
            "units": "0.3048 metre",
        },
        {
            "axis": "X",
            "long_name": "Easting",
            "standard_name": "projection_x_coordinate",
            "units": "0.3048 metre",
        },
    ]


def test_cartesian2dcs_to_cf():
    csm = Cartesian2DCS(axis=Cartesian2DCSAxis.EASTING_NORTHING_FT)
    csm.to_cf() == [
        {
            "axis": "Y",
            "long_name": "Northing",
            "standard_name": "projection_y_coordinate",
            "units": "metre",
        },
        {
            "axis": "X",
            "long_name": "Easting",
            "standard_name": "projection_x_coordinate",
            "units": "metre",
        },
    ]


def test_verticalcs_depth_to_cf():
    vcs = VerticalCS(axis=VerticalCSAxis.DEPTH)
    vcs.to_cf() == [
        {
            "standard_name": "height_above_reference_ellipsoid",
            "long_name": "Depth",
            "units": "metre",
            "positive": "down",
            "axis": "Z",
        }
    ]


def test_verticalcs_height_to_cf():
    vcs = VerticalCS(axis=VerticalCSAxis.GRAVITY_HEIGHT_US_FT)
    vcs.to_cf() == [
        {
            "standard_name": "height_above_reference_ellipsoid",
            "long_name": "Gravity-related height",
            "units": "0.304800609601219 metre",
            "positive": "up",
            "axis": "Z",
        }
    ]



import array

import numpy
from numpy.testing import assert_allclose

from pyproj import Proj, __proj_version__

try:
    from time import perf_counter
except ImportError:
    from time import clock as perf_counter


def test_awips221():
    params = {}
    params["proj"] = "lcc"
    params["R"] = 6371200
    params["lat_1"] = 50
    params["lat_2"] = 50
    params["lon_0"] = -107
    nx = 349
    ny = 277
    dx = 32463.41
    dy = dx
    # can either use a dict
    # awips221 = Proj(params)
    # or keyword args
    awips221 = Proj(proj="lcc", R=6371200, lat_1=50, lat_2=50, lon_0=-107)
    print("proj4 library version = ", __proj_version__)
    # AWIPS grid 221 parameters
    # (from http://www.nco.ncep.noaa.gov/pmb/docs/on388/tableb.html)
    llcrnrx, llcrnry = awips221(-145.5, 1.0)
    params["x_0"] = -llcrnrx
    params["y_0"] = -llcrnry
    awips221 = Proj(params)
    llcrnrx, llcrnry = awips221(-145.5, 1.0)
    # find 4 lon/lat crnrs of AWIPS grid 221.
    llcrnrx = 0.0
    llcrnry = 0.0
    lrcrnrx = dx * (nx - 1)
    lrcrnry = 0.0
    ulcrnrx = 0.0
    ulcrnry = dy * (ny - 1)
    urcrnrx = dx * (nx - 1)
    urcrnry = dy * (ny - 1)
    llcrnrlon, llcrnrlat = awips221(llcrnrx, llcrnry, inverse=True)
    lrcrnrlon, lrcrnrlat = awips221(lrcrnrx, lrcrnry, inverse=True)
    urcrnrlon, urcrnrlat = awips221(urcrnrx, urcrnry, inverse=True)
    ulcrnrlon, ulcrnrlat = awips221(ulcrnrx, ulcrnry, inverse=True)
    print("4 crnrs of AWIPS grid 221:")
    print(llcrnrlon, llcrnrlat)
    print(lrcrnrlon, lrcrnrlat)
    print(urcrnrlon, urcrnrlat)
    print(ulcrnrlon, ulcrnrlat)
    print("from GRIB docs")
    print("(see http://www.nco.ncep.noaa.gov/pmb/docs/on388/tableb.html)")
    print("   -145.5  1.0")
    print("   -68.318 0.897")
    print("   -2.566 46.352")
    print("   148.639 46.635")
    # compute lons and lats for the whole AWIPS grid 221 (377x249).
    dx = (urcrnrx - llcrnrx) / (nx - 1)
    dy = (urcrnry - llcrnry) / (ny - 1)
    x = llcrnrx + dx * numpy.indices((ny, nx), "f")[1, :, :]
    y = llcrnry + dy * numpy.indices((ny, nx), "f")[0, :, :]
    t1 = perf_counter()
    lons, lats = awips221(
        numpy.ravel(x).tolist(), numpy.ravel(y).tolist(), inverse=True
    )
    t2 = perf_counter()
    print("data in lists:")
    print("compute lats/lons for all points on AWIPS 221 grid ({}x{})".format(nx, ny))
    print("max/min lons")
    print(min(lons), max(lons))
    print("max/min lats")
    print(min(lats), max(lats))
    print("took", t2 - t1, "secs")
    xa = array.array("f", numpy.ravel(x).tolist())
    ya = array.array("f", numpy.ravel(y).tolist())
    t1 = perf_counter()
    lons, lats = awips221(xa, ya, inverse=True)
    t2 = perf_counter()
    print("data in python arrays:")
    print("compute lats/lons for all points on AWIPS 221 grid ({}x{})".format(nx, ny))
    print("max/min lons")
    print(min(lons), max(lons))
    print("max/min lats")
    print(min(lats), max(lats))
    print("took", t2 - t1, "secs")
    t1 = perf_counter()
    lons, lats = awips221(x, y, inverse=True)
    t2 = perf_counter()
    print("data in a numpy array:")
    print("compute lats/lons for all points on AWIPS 221 grid ({}x{})".format(nx, ny))
    print("max/min lons")
    print(
        numpy.minimum.reduce(numpy.ravel(lons)), numpy.maximum.reduce(numpy.ravel(lons))
    )
    print("max/min lats")
    print(
        numpy.minimum.reduce(numpy.ravel(lats)), numpy.maximum.reduce(numpy.ravel(lats))
    )
    print("took", t2 - t1, "secs")
    # reverse transformation.
    t1 = perf_counter()
    xx, yy = awips221(lons, lats)
    t2 = perf_counter()
    print("max abs error for x")
    max_abs_err_x = numpy.maximum.reduce(numpy.fabs(numpy.ravel(x - xx)))
    print(max_abs_err_x)
    assert_allclose(max_abs_err_x, 0, atol=1e-4)
    print("max abs error for y")
    max_abs_err_y = numpy.maximum.reduce(numpy.fabs(numpy.ravel(y - yy)))
    print(max_abs_err_y)
    assert_allclose(max_abs_err_y, 0, atol=1e-4)
    print("took", t2 - t1, "secs")
    print("compare output with sample.out")



import numpy
import pytest
from numpy.testing import assert_allclose, assert_almost_equal

from pyproj import Proj, __proj_version__, transform
from test.conftest import PROJ_911, PROJ_GTE_92, PROJ_GTE_911, grids_available


def test_transform():
    # convert awips221 grid to awips218 coordinate system
    # (grids defined at http://www.nco.ncep.noaa.gov/pmb/docs/on388/tableb.html)
    nx = 614
    ny = 428
    dx = 12190.58
    dy = dx
    awips221 = Proj(proj="lcc", R=6371200, lat_1=50, lat_2=50, lon_0=-107)
    print("proj4 library version = ", __proj_version__)
    llcrnrx, llcrnry = awips221(-145.5, 1)
    awips221 = Proj(
        proj="lcc",
        R=6371200,
        lat_1=50,
        lat_2=50,
        lon_0=-107,
        x_0=-llcrnrx,
        y_0=-llcrnry,
    )
    assert_allclose(awips221(-145.5, 1), (0, 0), atol=1e-4)
    awips218 = Proj(proj="lcc", R=6371200, lat_1=25, lat_2=25, lon_0=-95)
    llcrnrx, llcrnry = awips218(-133.459, 12.19)
    awips218 = Proj(
        proj="lcc", R=6371200, lat_1=25, lat_2=25, lon_0=-95, x_0=-llcrnrx, y_0=-llcrnry
    )
    assert_allclose(awips218(-133.459, 12.19), (0, 0), atol=1e-4)
    x1 = dx * numpy.indices((ny, nx), "f")[1, :, :]
    y1 = dy * numpy.indices((ny, nx), "f")[0, :, :]
    print("max/min x and y for awips218 grid")
    print(numpy.minimum.reduce(numpy.ravel(x1)), numpy.maximum.reduce(numpy.ravel(x1)))
    print(numpy.minimum.reduce(numpy.ravel(y1)), numpy.maximum.reduce(numpy.ravel(y1)))
    with pytest.warns(FutureWarning):
        x2, y2 = transform(awips218, awips221, x1, y1)
    print("max/min x and y for awips218 grid in awips221 coordinates")
    print(numpy.minimum.reduce(numpy.ravel(x2)), numpy.maximum.reduce(numpy.ravel(x2)))
    print(numpy.minimum.reduce(numpy.ravel(y2)), numpy.maximum.reduce(numpy.ravel(y2)))
    with pytest.warns(FutureWarning):
        x3, y3 = transform(awips221, awips218, x2, y2)
    print("error for reverse transformation back to awips218 coords")
    print("(should be close to zero)")
    assert_allclose(numpy.minimum.reduce(numpy.ravel(x3 - x1)), 0, atol=1e-4)
    assert_allclose(numpy.maximum.reduce(numpy.ravel(x3 - x1)), 0, atol=1e-4)
    assert_allclose(numpy.minimum.reduce(numpy.ravel(y3 - y1)), 0, atol=1e-4)
    assert_allclose(numpy.maximum.reduce(numpy.ravel(y3 - y1)), 0, atol=1e-4)


@pytest.mark.xfail(PROJ_911, reason="Patch applied in conda-forge changes behavior")
def test_transform_single_point_nad83_to_nad27():
    # projection 1: UTM zone 15, grs80 ellipse, NAD83 datum
    # (defined by epsg code 26915)
    p1 = Proj("epsg:26915", preserve_units=False)
    # projection 2: UTM zone 15, clrk66 ellipse, NAD27 datum
    p2 = Proj("epsg:26715", preserve_units=False)
    # find x,y of Jefferson City, MO.
    x1, y1 = p1(-92.199881, 38.56694)
    # transform this point to projection 2 coordinates.
    x2, y2 = transform(p1, p2, x1, y1)
    assert_almost_equal(
        (x1, y1),
        (569704.566, 4269024.671),
        decimal=3,
    )
    expected_xy2 = (569722, 4268814)
    if PROJ_GTE_911:
        expected_xy2 = (569720, 4268813)
        if (
            PROJ_GTE_92
            and grids_available(
                "us_noaa_nadcon5_nad27_nad83_1986_conus.tif", check_network=False
            )
        ) or grids_available():
            expected_xy2 = (569722, 4268814)
        elif grids_available(
            "ca_nrc_ntv2_0.tif", "ca_nrc_ntv1_can.tif", check_network=False
        ):
            expected_xy2 = (569706, 4268817)
        elif grids_available("us_noaa_conus.tif", check_network=False):
            expected_xy2 = (569722, 4268814)

    assert_almost_equal(
        (x2, y2),
        expected_xy2,
        decimal=0,
    )
    assert_almost_equal(
        p2(x2, y2, inverse=True, errcheck=True),
        (-92.200, 38.567),
        decimal=3,
    )


@pytest.mark.xfail(PROJ_911, reason="Patch applied in conda-forge changes behavior")
def test_transform_tuple_nad83_to_nad27():
    # projection 1: UTM zone 15, grs80 ellipse, NAD83 datum
    # (defined by epsg code 26915)
    p1 = Proj("epsg:26915", preserve_units=False)
    # projection 2: UTM zone 15, clrk66 ellipse, NAD27 datum
    p2 = Proj("epsg:26715", preserve_units=False)
    # process 3 points at a time in a tuple
    lats = (38.83, 39.32, 38.75)  # Columbia, KC and StL Missouri
    lons = (-92.22, -94.72, -90.37)
    x1, y1 = p1(lons, lats)
    x2, y2 = transform(p1, p2, x1, y1)
    assert_almost_equal(
        x1,
        (567703.344, 351730.944, 728553.093),
        decimal=3,
    )
    assert_almost_equal(
        y1,
        (4298200.739, 4353698.725, 4292319.005),
        decimal=3,
    )
    expected_x2 = (567721, 351747, 728569)
    expected_y2 = (4297989, 4353489, 4292106)
    if PROJ_GTE_911:
        expected_x2 = (567719, 351748, 728568)
        expected_y2 = (4297989, 4353487, 4292108)
        if (
            PROJ_GTE_92
            and grids_available(
                "us_noaa_nadcon5_nad27_nad83_1986_conus.tif", check_network=False
            )
        ) or grids_available():
            expected_x2 = (567721, 351747, 728569)
            expected_y2 = (4297989, 4353489, 4292106)
        elif grids_available(
            "ca_nrc_ntv2_0.tif", "ca_nrc_ntv1_can.tif", check_network=False
        ):
            expected_x2 = (567705, 351727, 728558)
            expected_y2 = (4297993, 4353490, 4292111)
        elif grids_available("us_noaa_conus.tif", check_network=False):
            expected_x2 = (567721, 351747, 728569.133)
            expected_y2 = (4297989, 4353489, 4292106)

    assert_almost_equal(
        x2,
        expected_x2,
        decimal=0,
    )
    assert_almost_equal(
        y2,
        expected_y2,
        decimal=0,
    )
    lons2, lats2 = p2(x2, y2, inverse=True, errcheck=True)
    assert_almost_equal(
        lons2,
        (-92.220, -94.720, -90.370),
        decimal=3,
    )
    assert_almost_equal(
        lats2,
        (38.830, 39.320, 38.750),
        decimal=3,
    )



include "proj.pxi"

from pyproj._datadir cimport PYPROJ_GLOBAL_CONTEXT


def get_proj_endpoint() -> str:
    """
    Returns
    -------
    str:
        URL of the endpoint where PROJ grids are stored.
    """
    return proj_context_get_url_endpoint(PYPROJ_GLOBAL_CONTEXT)



import argparse
import os
import subprocess
import sys
from glob import glob
from pathlib import Path
from unittest.mock import patch

import pytest

from pyproj.__main__ import main
from pyproj.datadir import append_data_dir, get_data_dir, get_user_data_dir
from pyproj.sync import _load_grid_geojson
from test.conftest import grids_available, proj_env, tmp_chdir

PYPROJ_CLI_ENDPONTS = pytest.mark.parametrize(
    "input_command", [["pyproj"], [sys.executable, "-m", "pyproj"]]
)


@pytest.mark.cli
@PYPROJ_CLI_ENDPONTS
def test_main(input_command, tmpdir):
    with tmp_chdir(str(tmpdir)):
        output = subprocess.check_output(
            input_command, stderr=subprocess.STDOUT
        ).decode("utf-8")
    assert "pyproj version:" in output
    assert "PROJ version:" in output
    assert "-v, --verbose  Show verbose debugging version information." in output


@pytest.mark.cli
@PYPROJ_CLI_ENDPONTS
@pytest.mark.parametrize("option", ["-v", "--verbose"])
def test_main__verbose(input_command, option, tmpdir):
    with tmp_chdir(str(tmpdir)):
        output = subprocess.check_output(
            input_command + [option], stderr=subprocess.STDOUT
        ).decode("utf-8")
    assert "pyproj:" in output
    assert "PROJ:" in output
    assert "data dir" in output
    assert "user_data_dir" in output
    assert "System" in output
    assert "python" in output
    assert "Python deps" in output
    assert "-v, --verbose " not in output


@pytest.mark.cli
@PYPROJ_CLI_ENDPONTS
@pytest.mark.parametrize("option", [["-h"], []])
def test_sync(input_command, option, tmpdir):
    with tmp_chdir(str(tmpdir)):
        output = subprocess.check_output(
            input_command + ["sync"] + option, stderr=subprocess.STDOUT
        ).decode("utf-8")
    assert (
        "Tool for synchronizing PROJ datum and transformation support data." in output
    )
    assert "--bbox" in output
    assert "--spatial-test" in output
    assert "--source-id" in output
    assert "--area-of-use" in output
    assert "--file" in output
    assert "--exclude-world-coverage" in output
    assert "--include-already-downloaded" in output
    assert "--list-files" in output
    assert "--system-directory" in output
    assert "--target-directory" in output
    assert "-v, --verbose" in output


def _check_list_files_header(lines):
    assert lines[0].rstrip("\r") == "filename | source_id | area_of_use"
    assert lines[1].rstrip("\r") == "----------------------------------"


@pytest.mark.cli
@pytest.mark.network
@PYPROJ_CLI_ENDPONTS
def test_sync__source_id__list(input_command, tmpdir):
    with tmp_chdir(str(tmpdir)):
        output = subprocess.check_output(
            input_command
            + [
                "sync",
                "--source-id",
                "fr_ign",
                "--list-files",
                "--include-already-downloaded",
            ],
            stderr=subprocess.STDOUT,
        ).decode("utf-8")
    lines = output.strip().split("\n")
    assert len(lines) > 2
    _check_list_files_header(lines)
    for line in lines[2:]:
        assert "fr_ign" == line.split("|")[1].strip()


@pytest.mark.cli
@pytest.mark.network
@PYPROJ_CLI_ENDPONTS
def test_sync__area_of_use__list(input_command, tmpdir):
    with tmp_chdir(str(tmpdir)):
        output = subprocess.check_output(
            input_command
            + [
                "sync",
                "--area-of-use",
                "France",
                "--list-files",
                "--include-already-downloaded",
            ],
            stderr=subprocess.STDOUT,
        ).decode("utf-8")
    lines = output.strip().split("\n")
    assert len(lines) > 2
    _check_list_files_header(lines)
    for line in lines[2:]:
        assert "France" in line.split("|")[-1]


@pytest.mark.cli
@pytest.mark.network
@PYPROJ_CLI_ENDPONTS
def test_sync__file__list(input_command, tmpdir):
    with tmp_chdir(str(tmpdir)):
        output = subprocess.check_output(
            input_command
            + [
                "sync",
                "--file",
                "ntf_r93",
                "--list-files",
                "--include-already-downloaded",
            ],
            stderr=subprocess.STDOUT,
        ).decode("utf-8")
    lines = output.strip().split("\n")
    assert len(lines) > 2
    _check_list_files_header(lines)
    for line in lines[2:]:
        assert "ntf_r93" in line.split("|")[0]


@pytest.mark.cli
@pytest.mark.network
@PYPROJ_CLI_ENDPONTS
def test_sync__bbox__list(input_command, tmpdir):
    with tmp_chdir(str(tmpdir)):
        output = subprocess.check_output(
            input_command
            + [
                "sync",
                "--bbox",
                "2,49,3,50",
                "--list-files",
                "--include-already-downloaded",
            ],
            stderr=subprocess.STDOUT,
        ).decode("utf-8")
    lines = output.strip().split("\n")
    assert len(lines) > 2
    _check_list_files_header(lines)
    assert " | be_ign | " in output
    assert " | us_nga | " in output
    assert " | fr_ign | " in output


@pytest.mark.cli
@pytest.mark.network
@PYPROJ_CLI_ENDPONTS
def test_sync__bbox__list__exclude_world_coverage(input_command, tmpdir):
    with tmp_chdir(str(tmpdir)):
        output = subprocess.check_output(
            input_command
            + [
                "sync",
                "--bbox",
                "2,49,3,50",
                "--exclude-world-coverage",
                "--list-files",
                "--include-already-downloaded",
            ],
            stderr=subprocess.STDOUT,
        ).decode("utf-8")
    lines = output.strip().split("\n")
    assert len(lines) > 2
    _check_list_files_header(lines)
    assert " | be_ign | " in output
    assert " | us_nga | " not in output
    assert " | fr_ign | " in output


@pytest.mark.cli
@PYPROJ_CLI_ENDPONTS
@pytest.mark.parametrize(
    "extra_arg",
    [
        "--list-files",
        "--source-id",
        "--area-of-use",
        "--bbox",
        "--list-files",
        "--file",
    ],
)
def test_sync__all__exclusive_error(input_command, extra_arg, tmpdir):
    with tmp_chdir(str(tmpdir)), pytest.raises(subprocess.CalledProcessError):
        subprocess.check_output(
            input_command + ["sync", "--all", extra_arg], stderr=subprocess.STDOUT
        )


@pytest.mark.network
@patch(
    "pyproj.__main__.parser.parse_args",
    return_value=argparse.Namespace(
        bbox=None,
        list_files=False,
        file="ntf_r93",
        all=False,
        source_id=None,
        area_of_use=None,
        verbose=False,
        target_directory=None,
        system_directory=False,
        spatial_test="intersects",
        exclude_world_coverage=False,
        include_already_downloaded=True,
    ),
)
@patch("pyproj.__main__._download_resource_file")
def test_sync_download(download_mock, parse_args_mock):
    main()
    download_mock.assert_called_with(
        directory=get_user_data_dir(),
        file_url="https://cdn.proj.org/fr_ign_ntf_r93.tif",
        sha256="0aa738b3e00fd2d64f8e3cd0e76034d4792374624fa0e133922433c9491bbf2a",
        short_name="fr_ign_ntf_r93.tif",
        verbose=False,
    )


@pytest.mark.network
@patch(
    "pyproj.__main__.parser.parse_args",
    return_value=argparse.Namespace(
        bbox=None,
        list_files=False,
        file="ntf_r93",
        all=False,
        source_id=None,
        area_of_use=None,
        verbose=True,
        target_directory="test_directory",
        system_directory=False,
        spatial_test="intersects",
        exclude_world_coverage=False,
        include_already_downloaded=True,
    ),
)
@patch("pyproj.__main__._download_resource_file")
@patch("pyproj.sync._load_grid_geojson")
def test_sync_download__directory(
    load_grid_geojson_mock, download_mock, parse_args_mock
):
    load_grid_geojson_mock.return_value = _load_grid_geojson()
    main()
    download_mock.assert_called_with(
        directory="test_directory",
        file_url="https://cdn.proj.org/fr_ign_ntf_r93.tif",
        sha256="0aa738b3e00fd2d64f8e3cd0e76034d4792374624fa0e133922433c9491bbf2a",
        short_name="fr_ign_ntf_r93.tif",
        verbose=True,
    )
    load_grid_geojson_mock.assert_called_with(target_directory="test_directory")


@pytest.mark.network
@patch(
    "pyproj.__main__.parser.parse_args",
    return_value=argparse.Namespace(
        bbox=None,
        list_files=False,
        file="ntf_r93",
        all=False,
        source_id=None,
        area_of_use=None,
        verbose=True,
        target_directory=None,
        system_directory=True,
        spatial_test="intersects",
        exclude_world_coverage=False,
        include_already_downloaded=True,
    ),
)
@patch("pyproj.__main__._download_resource_file")
@patch("pyproj.sync._load_grid_geojson")
def test_sync_download__system_directory(
    load_grid_geojson_mock, download_mock, parse_args_mock
):
    load_grid_geojson_mock.return_value = _load_grid_geojson()
    main()
    datadir = get_data_dir().split(os.path.sep)[0]
    download_mock.assert_called_with(
        directory=datadir,
        file_url="https://cdn.proj.org/fr_ign_ntf_r93.tif",
        sha256="0aa738b3e00fd2d64f8e3cd0e76034d4792374624fa0e133922433c9491bbf2a",
        short_name="fr_ign_ntf_r93.tif",
        verbose=True,
    )
    load_grid_geojson_mock.assert_called_with(target_directory=datadir)


@pytest.mark.network
@patch("pyproj.__main__.parser.parse_args")
def test_sync__download_grids(parse_args_mock, tmp_path, capsys):
    parse_args_mock.return_value = argparse.Namespace(
        bbox=None,
        list_files=False,
        file="us_noaa_alaska",
        all=False,
        source_id=None,
        area_of_use=None,
        verbose=True,
        target_directory=str(tmp_path),
        system_directory=False,
        spatial_test="intersects",
        exclude_world_coverage=False,
        include_already_downloaded=False,
    )
    main()
    captured = capsys.readouterr()
    paths = sorted(Path(path).name for path in glob(str(tmp_path.joinpath("*"))))
    if grids_available("us_noaa_alaska.tif", check_network=False):
        assert paths == ["files.geojson"]
        assert captured.out == ""
    else:
        assert paths == ["files.geojson", "us_noaa_alaska.tif"]
        assert captured.out == "Downloading: https://cdn.proj.org/us_noaa_alaska.tif\n"
    # make sure not downloaded again
    with proj_env():
        append_data_dir(str(tmp_path))
        main()
        captured = capsys.readouterr()
        assert captured.out == ""
